"""Constructing static objects.

Author(s): Erwin de Gelder
"""

from abc import ABC
from typing import Dict, Optional, Tuple, Union

import numpy as np
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon as PPolygon
from matplotlib.text import Text

from .options import Options
from .polygon import Polygon
from .utilities import rotate


class StaticObjectPosition(Options):
    """Parameters of a static object, containing the (x,y)-coordinates and the angle."""

    x_center: float = 0
    y_center: float = 0
    angle: float = 0


class StaticObject(ABC):
    """Abstract class for dealing with static objects.

    Attributes:
        axes (Axes): The axes object that is used to draw the sign.
        position (StaticObjectPosition): The (x,y)-coordinates of the center of the sign.
        fills (Tuple): A tuple containing the handles for the filled areas.
        plots (Tuple): A tuple containing the handles for the line plots.
        texts (Tuple): A tuple containing the handles for text objects.
    """

    axes: Axes
    fills: Tuple[Union[PPolygon, Polygon], ...]
    plots: Tuple[Line2D, ...]
    texts: Tuple[Text, ...]
    position: StaticObjectPosition

    def __init__(self, axes: Axes) -> None:
        """Creating a static object.

        :param axes: The axes on which the object is supposed to be drawn.
        """
        self.axes = axes
        self.fills = ()
        self.plots = ()
        self.texts = ()
        self.position = StaticObjectPosition()

    def change_pos(self, x_center: float, y_center: float, angle: float = 0) -> None:
        """Change the position of the static object.

        :param x_center: The new x-coordinate of the static object.
        :param y_center: The new y-coordinate of the static object.
        :param angle: The new angle of the static object.
        """
        for plot in self.plots:
            xdata = plot.get_xdata() - np.array(self.position.x_center)
            ydata = plot.get_ydata() - np.array(self.position.y_center)
            xdata, ydata = rotate(xdata, ydata, self.position.angle)
            xdata, ydata = rotate(xdata, ydata, -angle)
            plot.set_xdata(xdata + x_center)
            plot.set_ydata(ydata + y_center)

        for fill in self.fills:
            xy_data = fill.get_xy() - [self.position.x_center, self.position.y_center]
            xy_data[:, 0], xy_data[:, 1] = rotate(xy_data[:, 0], xy_data[:, 1], self.position.angle)
            xy_data[:, 0], xy_data[:, 1] = rotate(xy_data[:, 0], xy_data[:, 1], -angle)
            xy_data[:, 0] += x_center
            xy_data[:, 1] += y_center
            fill.set_xy(xy_data)

        for text in self.texts:
            text.set_position((x_center, y_center))

        self.position.x_center = x_center
        self.position.y_center = y_center
        self.position.angle = angle

    def change_color(
        self,
        face_color: Optional[Tuple[float, float, float]] = None,
        edge_color: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        """Change the colors of the filled areas and the plotted lines.

        :param face_color: The new color of the filled area.
        :param edge_color: The new color of the edge of the filled areas and the lines.
        """
        if face_color is not None:
            for fill in self.fills:
                fill.set_facecolor(face_color)
        if edge_color is not None:
            for fill in self.fills:
                fill.set_edgecolor(edge_color)
            for plot in self.plots:
                plot.set_color(edge_color)


class MaxSpeedOptions(Options):
    """Class containing the default values of the options of a way object."""

    fontsize: Optional[float] = None
    outer_radius: Optional[float] = None
    inner_radius: Optional[float] = None
    outer_color: Tuple[float, float, float] = (1, 0, 0)
    inner_color: Tuple[float, float, float] = (1, 1, 1)


class MaxSpeed(StaticObject):
    """A max speed sign that is drawn on a given axes.

    Attributes:
        options (MaxSpeedOptions): Options that can be set for the sign.
    """

    def __init__(
        self, axes: Axes, text: Optional[str] = None, options: Optional[MaxSpeedOptions] = None
    ) -> None:
        """Creating a sign with the maximum allowable speed.

        :param axes: The axes on which the speed sign is supposed to be drawn.
        :param text: The text that should be places inside the sign.
        :param options: Additional options for the speed sign (see MaxSpeedOptions).
        """
        self.options = MaxSpeedOptions() if options is None else options
        self.set_radius(axes)
        StaticObject.__init__(self, axes)

        # Draw the sign
        theta = np.linspace(0, 2 * np.pi, 40)
        xouter = np.cos(theta) * self.options.outer_radius
        youter = np.sin(theta) * self.options.outer_radius
        xinner = np.cos(theta) * self.options.inner_radius
        yinner = np.sin(theta) * self.options.inner_radius
        self.position = StaticObjectPosition()
        self.fills = (
            axes.fill(
                xouter + self.position.x_center,
                youter + self.position.y_center,
                color=self.options.outer_color,
            )[0],
            axes.fill(
                xinner + self.position.x_center,
                yinner + self.position.y_center,
                color=self.options.inner_color,
            )[0],
        )

        # Show text
        if text is not None:
            self.texts = (
                axes.text(
                    self.position.x_center,
                    self.position.y_center,
                    text,
                    horizontalalignment="center",
                    verticalalignment="center",
                    fontsize=self.options.fontsize,
                ),
            )

    def set_radius(self, axes: Axes) -> None:
        """Set the radius of the sign if it is not already set.

        If both the inner radius and the outer radius is set, nothing is done. If only
        the outer radius is set, the inner radius will be set to 80 % of the outer
        radius. If only the inner radius is set, the outer radius will be set to 125 %
        of the inner radius. If both radii are not set, the outer radius is set to 3.3 %
        of the width of the axes and the inner radius is set to 80 % of the outer radius.

        :param axes: The axes that is used to plot the sign on.
        """
        if self.options.inner_radius is None and self.options.outer_radius is None:
            self.options.outer_radius = np.diff(axes.get_xlim())[0] * 0.033
            self.options.inner_radius = 0.8 * self.options.outer_radius
        elif self.options.inner_radius is None and self.options.outer_radius is not None:
            self.options.inner_radius = 0.8 * self.options.outer_radius
        elif self.options.outer_radius is None and self.options.inner_radius is not None:
            self.options.outer_radius = 1.25 * self.options.inner_radius


class TurnArrowOptions(Options):
    """Class containing the default values of the options of a turn arrow object."""

    width: float = 0.9
    length: float = 2.5
    face_color: Tuple[float, float, float] = (0.8, 0.8, 0.8)
    edge_color: Tuple[float, float, float] = (0, 0, 0)
    layer: int = 0


class TurnArrow(StaticObject):
    """A turning arrow that is typically drawn on the road.

    The turning arrow shows the supposed direction of vehicles driving in the
    specific lane. Currently, the directions 'left', 'right', 'leftright',
    'leftthrough', 'throughright', and 'through' are supported.

    Attributes:
        direction (str): The direction that is indicated by the turning arrow.
        options (TurnArrowOptions): Options that can be set for the sign.
    """

    def __init__(
        self,
        axes: Axes,
        direction: Optional[str] = None,
        options: Optional[TurnArrowOptions] = None,
    ) -> None:
        """Creating a turning arrow.

        :param axes: The axes on which the turning arrow is supposed to be drawn.
        :param direction: Choose between 'through' (default), 'left', 'right', 'leftright',
                         'leftthrough', and 'throughright'.
        :param options: Additional options for the turning arrow (see TurnArrowOptions).
        """
        self.options = TurnArrowOptions() if options is None else options
        StaticObject.__init__(self, axes)
        self.direction = direction

        # Get coordinates of the arrow
        if self.direction in ("right", "left"):
            xdata = [-0.3, 0.3, 0.12, 0.24, 0.42, 0.54, 0.54, 1.2, 0.54, 0.54, 0.3, 0.12, -0.12]
            ydata = [
                -0.5,
                -0.5,
                0.152,
                0.174,
                0.189,
                0.189,
                0.053,
                0.242,
                0.5,
                0.28,
                0.28,
                0.265,
                0.227,
            ]
            if self.direction == "left":
                xdata = [-xdata_element for xdata_element in xdata]
        elif self.direction == "leftright":
            xdata = [
                -0.3,
                0.3,
                0.12,
                0.24,
                0.42,
                0.54,
                0.54,
                1.2,
                0.54,
                0.54,
                0.3,
                0.12,
                -0.12,
                -0.3,
                -0.54,
                -0.54,
                -1.2,
                -0.54,
                -0.54,
                -0.42,
                -0.24,
                -0.12,
            ]
            ydata = [
                -0.5,
                -0.5,
                0.152,
                0.174,
                0.189,
                0.189,
                0.053,
                0.242,
                0.5,
                0.28,
                0.28,
                0.265,
                0.265,
                0.28,
                0.28,
                0.5,
                0.242,
                0.053,
                0.189,
                0.189,
                0.174,
                0.152,
            ]
        elif self.direction in ("leftthrough", "throughright"):
            xdata = [
                0.3,
                0.233,
                0.433,
                0.833,
                0.9,
                0.9,
                1.3,
                0.9,
                0.9,
                0.967,
                0.433,
                0.233,
                0.167,
                0.167,
                0.5,
                0,
                -0.5,
                -0.167,
                -0.3,
            ]
            ydata = [
                -0.5,
                -0.102,
                -0.078,
                -0.063,
                -0.063,
                -0.133,
                -0.023,
                0.102,
                0.023,
                0.023,
                0.016,
                0,
                0,
                0.117,
                0.117,
                0.5,
                0.117,
                0.117,
                -0.5,
            ]
            if self.direction == "leftthrough":
                xdata = [-xdata_element for xdata_element in xdata]
        else:
            xdata = [-0.3, 0.3, 0.167, 0.5, 0, -0.5, -0.167]
            ydata = [-0.5, -0.5, 0.105, 0.105, 0.5, 0.105, 0.105]

        # Set width and height
        xdata_array = np.concatenate((xdata, [xdata[0]])) * self.options.width
        ydata_array = np.concatenate((ydata, [ydata[0]])) * self.options.length

        # Plot the arrow
        self.position = StaticObjectPosition()
        self.fills = (
            axes.fill(
                xdata_array + self.position.x_center,
                ydata_array + self.position.y_center,
                facecolor=self.options.face_color,
                edgecolor=self.options.edge_color,
                zorder=self.options.layer,
            )[0],
        )


class BuildingOptions(Options):
    """Options for impassable objects."""

    size: float = 2
    size_x: Optional[float] = None
    size_y: Optional[float] = None
    x_data: Optional[np.ndarray] = None
    y_data: Optional[np.ndarray] = None
    face_color: Tuple[float, float, float] = (0.5, 0.5, 1)
    edge_color: Tuple[float, float, float] = (0, 0, 1)
    hatch: str = "//"


class Building(StaticObject):
    """A static object, e.g., a building, an impassable object, or a passable object."""

    def __init__(self, axes: Axes, options: Optional[BuildingOptions] = None) -> None:
        """Creating a 'building'.

        :param axes: The axes on which the building is supposed to be drawn.
        :param options: Additional options for the building (see BuildingOptions).
        """
        self.options = BuildingOptions() if options is None else options
        StaticObject.__init__(self, axes)

        if self.options.size_x is None:
            self.options.size_x = self.options.size
        if self.options.size_y is None:
            self.options.size_y = self.options.size

        if self.options.x_data is None:
            self.options.x_data = np.array([1, 1, -1, -1]) * self.options.size_x / 2
            self.options.y_data = np.array([1, -1, -1, 1]) * self.options.size_y / 2

        fill_options: Dict[str, Union[Tuple[float, float, float], str]]
        fill_options = {"facecolor": self.options.face_color, "edgecolor": self.options.edge_color}
        if self.options.hatch:
            fill_options["hatch"] = self.options.hatch
        self.fills = (axes.fill(self.options.x_data, self.options.y_data, **fill_options)[0],)


class StripesOptions(TurnArrowOptions):
    """Class containing the default values of the options of a stripes object."""

    face_color2: Tuple[float, float, float] = (1.0, 0.0, 0.0)
    nstripes: int = 10


class Stripes(StaticObject):
    """Class for drawing a rectangular object with diagonal stripes."""

    def __init__(self, axes: Axes, options: Optional[StripesOptions] = None) -> None:
        """Creating an object with diagonal stripes (e.g., for a road works sign).

        :param axes: The axes on which the turning arrow is supposed to be drawn.
        :param options: Additional options for the stripes object (see StripesOptions).
        """
        StaticObject.__init__(self, axes)
        self.options = StripesOptions() if options is None else options

        # Draw the rectangle.
        self.fills = (
            axes.fill(
                np.array([-1, 1, 1, -1]) * self.options.width / 2,
                np.array([1, 1, -1, -1]) * self.options.length / 2,
                facecolor=self.options.face_color,
                edgecolor=None,
                zorder=self.options.layer,
            )[0],
        )
        w_stripe = (self.options.width + self.options.length) / self.options.nstripes
        for i in range((self.options.nstripes + 1) // 2):
            alphai, betai = 2 * i * w_stripe, (2 * i + 1) * w_stripe

            xdata = [min(alphai, self.options.width)]
            ydata = [alphai - xdata[0]]

            if alphai < self.options.width < betai:
                xdata.append(self.options.width)
                ydata.append(0)

            xdata.append(min(betai, self.options.width))
            ydata.append(betai - xdata[-1])

            if i < self.options.nstripes - 1:
                ydata.append(min(betai, self.options.length))
                xdata.append(betai - ydata[-1])

            if alphai < self.options.length < betai:
                xdata.append(0)
                ydata.append(self.options.length)

            if i > 0:
                ydata.append(min(alphai, self.options.length))
                xdata.append(alphai - ydata[-1])

            self.fills += (
                axes.fill(
                    np.array(xdata) - self.options.width / 2,
                    self.options.length / 2 - np.array(ydata),
                    facecolor=self.options.face_color2,
                    edgecolor=None,
                    zorder=self.options.layer,
                )[0],
            )
        self.plots = (
            axes.plot(
                np.array([-1, 1, 1, -1, -1]) * self.options.width / 2,
                np.array([1, 1, -1, -1, 1]) * self.options.length / 2,
                color=self.options.edge_color,
                zorder=self.options.layer,
            )[0],
        )

    def change_color(
        self,
        face_color: Optional[Tuple[float, float, float]] = None,
        edge_color: Optional[Tuple[float, float, float]] = None,
        face_color2: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        """Change the colors of the filled areas and the plotted lines.

        :param face_color: The new color of the filled area.
        :param edge_color: The new color of the edge of the filled areas and the lines.
        :param face_color2: The new color of the stripes.
        """
        if face_color is not None:
            self.fills[0].set_facecolor(face_color)
        if edge_color is not None:
            self.plots[0].set_color(edge_color)
        if face_color2 is not None:
            for fill in self.fills[1:]:
                fill.set_facecolor(face_color2)



"""Class that can be used to make an object follow a certain path.

Author(s): Erwin de Gelder
"""

from typing import Optional, Tuple

import numpy as np


class PathFollowerLengthNotSetError(Exception):
    """Error to be raised in case the length is not set."""

    def __init__(self) -> None:
        """Description of error."""
        super().__init__("Length of PathFollower not set.")


class PathFollower:
    """Class for making an object follow a path."""

    def __init__(
        self,
        xcoordinates: np.ndarray,
        ycoordinates: np.ndarray,
        length: Optional[float] = None,
        init_rear: Optional[Tuple[int, float]] = None,
    ) -> None:
        """Create a PathFollower object that can be used to make an object follow a path.

        The location of the rear of the vehicle will be represented using (i_segment, lambda),
        where i_segment represents the segment the rear is on. Note that there are N-1 segments,
        where N represents the number of elements in the x,y coordinates. The proportion of the
        segment that is already covered is denoted by lambda.

        :param xcoordinates: The x-coordinates of the path.
        :param ycoordinates: The y-coordinates of the path.
        :param length: The length of the vehicle. It may be set later.
        :param init_rear: Initial location of the rear vehicle represented using
                          (i_segment, lambda). By default, this is (0, 0).
        """
        self.xcoordinates = xcoordinates
        self.ycoordinates = ycoordinates
        self.length = length
        if init_rear is None:
            self.i_segment = 0
            self.lambda_segment = 0.0
        else:
            self.i_segment, self.lambda_segment = init_rear
        self._segment_lengths = np.hypot(np.diff(self.xcoordinates), np.diff(self.ycoordinates))

    def get_rear_xy(self) -> Tuple[float, float]:
        """Return the (x,y) coordinates of the rear of the vehicle.

        :return: [x-coordinate, y-coordinate]
        """
        return (
            self.xcoordinates[self.i_segment] * (1 - self.lambda_segment)
            + self.xcoordinates[self.i_segment + 1] * self.lambda_segment,
            self.ycoordinates[self.i_segment] * (1 - self.lambda_segment)
            + self.ycoordinates[self.i_segment + 1] * self.lambda_segment,
        )

    def get_location_front(self) -> Tuple[int, float]:
        """Find the location of the front of the vehicle, given its position of the rear.

        :return: [Segment of front, Fraction of segment covered by front]
        """
        # Check that length is set.
        if self.length is None:
            raise PathFollowerLengthNotSetError

        # First determine the segment that the front is located in.
        xrear, yrear = self.get_rear_xy()
        i = self.i_segment + 1
        while np.hypot(self.xcoordinates[i] - xrear, self.ycoordinates[i] - yrear) < self.length:
            i += 1

        # We can now find lambda using quadratic solving. To see how this works, just write down the
        # equations yourself.
        # Also, except in some weird cases (e.g., if the path is screwed up), we can simply take the
        # largest solution.
        quadratic_a = (self.xcoordinates[i] - self.xcoordinates[i - 1]) ** 2 + (
            self.ycoordinates[i] - self.ycoordinates[i - 1]
        ) ** 2
        quadratic_b = 2 * (
            (self.xcoordinates[i - 1] - xrear) * (self.xcoordinates[i] - self.xcoordinates[i - 1])
            + (self.ycoordinates[i - 1] - yrear) * (self.ycoordinates[i] - self.ycoordinates[i - 1])
        )
        quadratic_c = (
            (self.xcoordinates[i - 1] - xrear) ** 2
            + (self.ycoordinates[i - 1] - yrear) ** 2
            - self.length**2
        )
        lambda_segment = (
            -quadratic_b + np.sqrt(quadratic_b**2 - 4 * quadratic_a * quadratic_c)
        ) / (2 * quadratic_a)
        return i, lambda_segment

    def get_front_xy(self) -> Tuple[float, float]:
        """Return the (x,y) coordinates of the front of the vehicle.

        :return: [x-coordinate, y-coordinate]
        """
        i_segment, lambda_segment = self.get_location_front()
        return (
            self.xcoordinates[i_segment - 1] * (1 - lambda_segment)
            + self.xcoordinates[i_segment] * lambda_segment,
            self.ycoordinates[i_segment - 1] * (1 - lambda_segment)
            + self.ycoordinates[i_segment] * lambda_segment,
        )

    def get_center_coordinates(self) -> Tuple[float, float, float]:
        """Get the center coordinates, including the heading.

        :return: [x-coordinate, y-coordinate, angle]
        """
        xrear, yrear = self.get_rear_xy()
        xfront, yfront = self.get_front_xy()
        angle = np.arctan2(xfront - xrear, yfront - yrear)
        return (xfront + xrear) / 2, (yfront + yrear) / 2, angle

    def move_vehicle(self, stepsize: float) -> Tuple[float, float, float]:
        """Move the vehicle a tiny bit and return new coordinates.

        :return: [x-coordinate, y-coordinate, angle]
        """
        # Check if we can use the same segment.
        available_space = (1 - self.lambda_segment) * self._segment_lengths[self.i_segment]
        if available_space > stepsize:
            self.lambda_segment += stepsize / self._segment_lengths[self.i_segment]
        else:
            stepsize -= available_space

            # Find the right segment
            self.i_segment += 1
            while stepsize > self._segment_lengths[self.i_segment]:
                stepsize -= self._segment_lengths[self.i_segment]
                self.i_segment += 1

            # Compute the new lambda
            self.lambda_segment = stepsize / self._segment_lengths[self.i_segment]

        # Return the updated position
        return self.get_center_coordinates()



"""Scripts for testing road_network.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path
from typing import Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from traffic_scene_renderer import (
    RoadNetwork,
    RoadNetworkOptions,
    StopLineOptions,
    Vertex,
    Way,
    WayOptions,
)

from .test_static_objects import save_fig

mpl.use("Agg")


def test_road_network_creation() -> None:
    vertices = [Vertex(1, -10, 0), Vertex(2, 10, 0)]
    ways = [Way(vertices)]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 5)
    save_fig(fig, axes, Path("road_network") / "straight_road.png", 10)


def test_automatic_crossing_creation() -> None:
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0), Vertex(3, 0, 10)]
    ways = [Way(vertices[:3]), Way([vertices[1], vertices[3]])]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 10)
    save_fig(fig, axes, Path("road_network") / "crossing.png", 10)


def test_automatic_connection_creation() -> None:
    vertices = [
        Vertex(0, -10, 10),
        Vertex(1, -10, 0),
        Vertex(2, 0, 0),
        Vertex(3, 10, 0),
        Vertex(4, 10, 10),
    ]
    ways = [
        Way([vertices[0], vertices[1]], WayOptions(side_color=(1, 0, 0))),
        Way([vertices[2], vertices[1]], WayOptions(side_color=(0, 0, 1))),
        Way([vertices[3], vertices[2]], WayOptions(side_color=(1, 0, 0))),
        Way([vertices[3], vertices[4]], WayOptions(side_color=(0, 0, 1))),
    ]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-14, 14)
    axes.set_ylim(-5, 10)
    save_fig(fig, axes, Path("road_network") / "connections.png", 14)


def test_check_crossing_reprocessing() -> None:
    vertices = [
        Vertex(0, -20, 0),
        Vertex(1, -10, 0),
        Vertex(2, -10, 10),
        Vertex(3, 0, 0),
        Vertex(4, 10, 0),
        Vertex(5, 10, 10),
        Vertex(6, 20, 0),
    ]
    ways = [
        Way([vertices[0], vertices[1]]),
        Way([vertices[1], vertices[2]]),
        Way([vertices[1], vertices[3]]),
        Way([vertices[3], vertices[4]]),
        Way([vertices[4], vertices[5]]),
        Way([vertices[4], vertices[6]]),
    ]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-20, 20)
    axes.set_ylim(-5, 10)
    save_fig(fig, axes, Path("road_network") / "crossing_and_connection.png", 20)


def test_stopline_not_added() -> None:
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, 0, 0),
        Vertex(2, 10, 0),
        Vertex(3, 20, 0),
        Vertex(4, 0, 10),
    ]
    ways = [
        Way(vertices[:-2]),
        Way(vertices[2:4]),
        Way([vertices[1], vertices[4]], WayOptions(oneway=True)),
    ]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, _ = road_network.plot()
    assert not road_network.add_stopline(ways[1])  # Way is not connected to crossing
    assert not road_network.add_stopline(ways[0], stoplineoptions=StopLineOptions(stopline=False))
    assert not road_network.add_stopline(ways[2])  # One way road that starts at crossing
    plt.close(fig)


def test_stopline() -> None:
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0), Vertex(3, 0, 10)]
    ways = [Way(vertices[:3]), Way([vertices[3], vertices[1]], WayOptions(oneway=True))]
    road_network = RoadNetwork(ways, vertices, RoadNetworkOptions(rightdrive=False))
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 10)
    assert road_network.add_stopline(ways[0])
    assert road_network.add_stopline(ways[1])
    save_fig(fig, axes, Path("road_network") / "stopline.png", 10)


def test_stopsign() -> None:
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0), Vertex(3, 0, 10)]
    ways = [Way(vertices[:3]), Way([vertices[3], vertices[1]])]
    road_network = RoadNetwork(ways, vertices, RoadNetworkOptions(rightdrive=True))
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 10)
    assert road_network.add_stopline(ways[1], stoplineoptions=StopLineOptions(stopsign=True))
    save_fig(fig, axes, Path("road_network") / "stopsign.png", 10)


def test_dirsign() -> None:
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0), Vertex(3, 0, 13)]
    ways = [Way(vertices[:3]), Way([vertices[3], vertices[1]])]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 13)
    assert road_network.add_stopline(
        ways[1], stoplineoptions=StopLineOptions(stopsign=True, dir_signs=["leftright", None])
    )
    save_fig(fig, axes, Path("road_network") / "dirsign.png", 10)


def simple_crossing() -> Tuple[Figure, Axes, RoadNetwork]:
    vertices = [
        Vertex(0, -10, 0),
        Vertex(1, 0, 0),
        Vertex(2, 10, 0),
        Vertex(3, 0, 10),
        Vertex(4, 0, -10),
    ]
    ways = [Way(vertices[:3]), Way([vertices[3], vertices[1], vertices[4]])]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, axes = road_network.plot()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-10, 10)
    return fig, axes, road_network


def test_trafficlights_leftright() -> None:
    fig, axes, road_network = simple_crossing()
    road_network.add_traffic_lights()
    save_fig(fig, axes, Path("road_network") / "traffic_lights_left_right.png", 10)


def test_trafficlights_none() -> None:
    fig, _, road_network = simple_crossing()
    assert not road_network.add_traffic_lights(leftright=(False, False))
    plt.close(fig)


def test_traffic_light_with_oneway_road() -> None:
    vertices = [Vertex(0, -10, 0), Vertex(1, 0, 0), Vertex(2, 10, 0), Vertex(3, 0, 10)]
    ways = [Way(vertices[:3]), Way([vertices[1], vertices[3]], WayOptions(nlanes=1, oneway=True))]
    road_network = RoadNetwork(ways, vertices, options=RoadNetworkOptions(rightdrive=False))
    road_network.process()
    fig, axes = road_network.plot()
    road_network.add_traffic_lights(leftright=(True, False))
    axes.set_xlim(-10, 10)
    axes.set_ylim(-5, 10)
    save_fig(fig, axes, Path("road_network") / "crossing.png", 10)


def test_traffic_light_without_crossing() -> None:
    vertices = [Vertex(0, 0, 0), Vertex(1, 10, 0)]
    ways = [Way(vertices)]
    road_network = RoadNetwork(ways, vertices)
    road_network.process()
    fig, _ = road_network.plot()
    assert not road_network.add_traffic_lights()
    plt.close(fig)


def test_traffic_light_with_stoplines() -> None:
    fig, axes, road_network = simple_crossing()
    road_network.add_traffic_lights(leftright=(False, True), stoplineoptions=StopLineOptions())
    plt.close(fig)



"""This file contains some simple functions that are useful but too small to have an own file.

Author(s): Erwin de Gelder
"""

from typing import Optional, Tuple

import numpy as np
import utm  # Installation required (pip install utm)


def wgs_to_utm(
    points_wgs: np.ndarray, force_zone_number: Optional[int] = None
) -> Tuple[np.ndarray, Optional[int], str]:
    """Convert WGS coordinates to UTM coordinates, such that zone is similar.

    :param points_wgs: N-by-2 array containing N lat-lon coordinates.
    :param force_zone_number: Zone number to be used. Set to none if zone number needs to be
                              determined by first datapoint (default: None).
    :return: A tuple with a N-by-2 array containing N easting-northing coordinates, an integer
             of the zone used for transformation, and the character of the zone.
    """
    n_points = points_wgs.shape[0]
    points_utm = np.empty((n_points, 2))
    zone_char = "U"  # Make sure it is defined

    for i in range(n_points):
        point_utm = utm.from_latlon(
            points_wgs[i, 0], points_wgs[i, 1], force_zone_number=force_zone_number
        )
        points_utm[i, 0] = point_utm[0]
        points_utm[i, 1] = point_utm[1]
        if i == 0:
            force_zone_number = point_utm[2]
            zone_char = point_utm[3]

    return points_utm, force_zone_number, zone_char


def rotate(x_data: np.ndarray, y_data: np.ndarray, angle: float) -> Tuple[np.ndarray, np.ndarray]:
    """Rotate the (x,y)-data around the origin by a specified angle.

    :param x_data: The x-coordinates of the data.
    :param y_data: The y-coordinates of the data.
    :param angle: The rotation angle.
    :return: A tuple containing the x-coordinates and the y-coordinates of the rotated data.
    """
    x_new = np.cos(angle) * x_data - np.sin(angle) * y_data
    y_new = np.sin(angle) * x_data + np.cos(angle) * y_data
    return x_new, y_new


def rgb2hsl(red: float, green: float, blue: float) -> Tuple[float, float, float]:
    """Convert RGB color format to HSL color format.

    :param red: Red content (value from 0 to 1).
    :param green: Green content (value from 0 to 1).
    :param blue: Blue content (value from 0 to 1).
    :return: Tuple with Hue, Saturation, and Luminance (values from 0 to 1).
    """
    cmax = max(red, green, blue)
    cmin = min(red, green, blue)
    delta = cmax - cmin

    if delta == 0.0:
        hue = 0.0
    elif cmax == red:
        hue = (green - blue) / (6 * delta) % 1
    elif cmax == green:
        hue = ((blue - red) / (2 * delta) + 1) / 3
    else:
        hue = ((red - green) / (2 * delta) + 2) / 3

    luminance = (cmax + cmin) / 2
    saturation = 0.0 if delta == 0.0 else delta / (1 - abs(2 * luminance - 1))

    return hue, saturation, luminance


def hsl2rgb(hue: float, saturation: float, luminance: float) -> Tuple[float, float, float]:
    """Convert HSL color format to RGB color format.

    :param hue: Hue value (value from 0 to 1).
    :param saturation: Saturation value (value from 0 to 1).
    :param luminance: Luminance value (value from 0 to 1).
    :return: Tuple with red, green, and blue content (values from 0 to 1).
    """
    first_color = (1 - abs(2 * luminance - 1)) * saturation
    second_color = first_color * (1 - abs((6 * hue % 2) - 1))
    mean_color = luminance - first_color / 2

    if hue < 1 / 6:
        red, green, blue = first_color, second_color, 0.0
    elif hue < 1 / 3:
        red, green, blue = second_color, first_color, 0.0
    elif hue < 1 / 2:
        red, green, blue = 0.0, first_color, second_color
    elif hue < 2 / 3:
        red, green, blue = 0.0, second_color, first_color
    elif hue < 5 / 6:
        red, green, blue = second_color, 0.0, first_color
    else:
        red, green, blue = first_color, 0.0, second_color

    red += mean_color
    green += mean_color
    blue += mean_color
    return red, green, blue



"""Scripts for testing polygon.py.

Author(s): Erwin de Gelder
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from traffic_scene_renderer.polygon import Polygon

mpl.use("Agg")
XDATA = np.array([0, 0, 1, 1])
YDATA = np.array([0, 1, 1, 0])


def test_polygon_creation() -> None:
    fig, axes = plt.subplots()
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon.remove()  # Remove it again
    plt.close(fig)


def test_polygon_get_xy_data() -> None:
    fig, axes = plt.subplots()
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=False)
    xdata = polygon.get_xdata()
    ydata = polygon.get_ydata()
    assert np.all(polygon.get_xy() == np.array([xdata, ydata]).T)
    plt.close(fig)


def test_polygon_set_xy_data() -> None:
    fig, axes = plt.subplots()
    polygon1 = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon1.set_xdata(XDATA + 1)
    polygon1.set_ydata(YDATA + 1)
    polygon2 = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon2.set_xy(np.array([XDATA + 1, YDATA + 1]).T)
    assert np.all(polygon1.get_xy() == polygon2.get_xy())
    plt.close(fig)


def test_polygon_set_colors() -> None:
    fig, axes = plt.subplots()
    # Check whether colors have changed
    my_colors = (0.1, 0.2, 0.3)
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon.set_color(my_colors)
    assert polygon.patch.get_edgecolor()[:3] == my_colors
    assert polygon.patch.get_facecolor()[:3] == my_colors

    # Check whether only facecolor have changed
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon.set_facecolor(my_colors)
    assert polygon.patch.get_edgecolor()[:3] != my_colors
    assert polygon.patch.get_facecolor()[:3] == my_colors

    # Check whether only edgecolor have changed
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=False)
    polygon.set_edgecolor(my_colors)
    assert polygon.patch.get_edgecolor()[:3] == my_colors
    assert polygon.patch.get_facecolor()[:3] != my_colors

    # Check whether none of the color have changed
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=True)
    polygon.set_color(my_colors)
    polygon.set_facecolor(my_colors)
    polygon.set_edgecolor(my_colors)
    assert polygon.patch.get_edgecolor()[:3] != my_colors
    assert polygon.patch.get_facecolor()[:3] != my_colors

    # Check whether only facecolor have changed
    polygon = Polygon(axes, XDATA, YDATA, fixed_color=True)
    polygon.set_facecolor_forced(my_colors)
    assert polygon.patch.get_edgecolor()[:3] != my_colors
    assert polygon.patch.get_facecolor()[:3] == my_colors
    plt.close(fig)



"""Constructing a road network.

Author(s): Erwin de Gelder
"""

from typing import List, NamedTuple, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from .connection import Connection
from .crossing import Crossing, CrossingOptions
from .letters import Letters, LettersOptions
from .options import Options
from .static_objects import TurnArrow, TurnArrowOptions
from .traffic_light import TrafficLight, TrafficLightOptions
from .vertex import Vertex
from .way import Way


class CrossingInfo(NamedTuple):
    """Information on crossings."""

    crossing: Crossing
    lam: float
    crossing_at_start: bool


class StopLineOptions(Options):
    """Options for a stop line."""

    stopline: bool = True  # By default, a stop line is added when calling the add_stopline.
    stopsign: bool = False  # By default, no stop sign is added.
    dir_signs: Optional[List[Union[str, None]]] = (
        None  # Whether to add directional signs on the road.
    )
    lonoffset: float = 0  # Further offset of the line from the crossing.


class RoadNetworkOptions(Options):
    """Options for a road network."""

    rightdrive: bool = True
    face_color: Tuple[float, float, float] = (0.8, 1, 0.8)
    face_color_roundabout: Optional[Tuple[float, float, float]] = None
    overrule_colors: bool = True
    line_color: Tuple[float, float, float] = (0, 0, 0)
    marker_border_color: Tuple[float, float, float] = (1, 1, 1)
    marker_fill_color: Tuple[float, float, float] = (1, 1, 1)


class RoadNetworkParameters(Options):
    """Parameters for a road network."""

    def __init__(
        self, **kwargs: Union[List[int], List[TrafficLight], List[TurnArrow], Axes, np.ndarray]
    ) -> None:
        """Initialize the parameters for a road network."""
        self.ivs: List[int] = []
        self.figure, self.axes = plt.subplots(1, 1, figsize=(15, 7))
        self.traffic_lights: List[TrafficLight] = []
        self.turn_arrows: List[TurnArrow] = []
        self.mat_edges: np.ndarray = np.array([])
        self.n_edges_per_vertex: np.ndarray = np.array([])
        Options.__init__(self, **kwargs)


class RoadNetwork:
    """Rendering a road network.

    Attributes:
        options(RoadNetworkOptions): Options for the road network.
        ways(List[Way]): The ways that are part of the road network.
        vertices(List[Vertex]): The vertices that are part of the road network.
        parms(RoadNetworkParameters): All kinds of parameters of the road network.
        crossings(List[Crossing]): The crossings that are part of the road network.
        connections(List[Connection]): The connections that are part of the road network.
    """

    def __init__(
        self, ways: List[Way], vertices: List[Vertex], options: Optional[RoadNetworkOptions] = None
    ) -> None:
        """Initialize a road network.

        :param ways: List of ways of the roadnetwork.
        :param vertices: List of vertices that are part of the network.
        :param options: Options for the road network, see RoadNetworkOptions.
        """
        self.options = RoadNetworkOptions() if options is None else options
        if self.options.face_color_roundabout is None:
            self.options.face_color_roundabout = self.options.face_color
        self.ways = ways
        self.vertices = vertices
        self.parms = RoadNetworkParameters()
        self.parms.ivs = [vertex.idx for vertex in self.vertices]

        # Compute number of edges.
        n_edges = 0
        for way in self.ways:
            n_edges += len(way.vertices) - 1

        # Create nv-by-ne boolean matrix (nv=#vertices, ne=#edges), where the (i,j)-th entry
        # is True if the i-th vertex belongs to the j-th edge.
        self.parms.mat_edges = np.zeros((len(self.vertices), n_edges), dtype=bool)
        i_edge = 0  # Index of edge.
        for way in self.ways:
            for i_vertex1, i_vertex2 in zip(way.ivs[:-1], way.ivs[1:]):
                self.parms.mat_edges[self.parms.ivs.index(i_vertex1), i_edge] = True
                self.parms.mat_edges[self.parms.ivs.index(i_vertex2), i_edge] = True
                i_edge += 1
        self.parms.n_edges_per_vertex = np.sum(self.parms.mat_edges, axis=1)

        self.split_ways()
        self.crossings = self.construct_crossings()
        self.connections = self.construct_connections()

    def split_ways(self) -> None:
        """Split ways, such that all inner vertices of a way have degree=2.

        A vertex that contains more than 2 edges while the vertex is not at the start or
        the end of a way, the way is split, such that this vertex is only at the start
        and end of a way.
        """
        n_ways = len(self.ways)
        i_way = 0
        while i_way < n_ways:  # A while loop, because number of ways can increase in loop.
            way = self.ways[i_way]
            for i_vertex in range(1, len(way.ivs) - 1):
                if self.parms.n_edges_per_vertex[way.ivs[i_vertex]] != 2:  # noqa: PLR2004
                    self.ways.append(way.split(i_vertex))
                    n_ways += 1
                    break
            i_way += 1

    def construct_crossings(self) -> List[Crossing]:
        """Construct the crossings of the road network.

        If a vertex contains more than 2 edges, a crossing is constructed.

        :return: List of crossings.
        """
        crossings = []
        n_crossings = 0
        for i, index_vertex in enumerate(self.parms.ivs):
            if self.parms.n_edges_per_vertex[i] > 2:  # noqa: PLR2004
                ways = [way for way in self.ways if index_vertex in way.ivs]
                crossings.append(
                    Crossing(
                        n_crossings,
                        self.vertices[i],
                        ways,
                        CrossingOptions(roundabout_inner_color=self.options.face_color_roundabout),
                    )
                )
                n_crossings += 1
        return crossings

    def construct_connections(self) -> List[Connection]:
        """Construct the connections of the road network.

        A connection connects ways that are not connected using a crossing. This is the case
        if a vertex at the start or end of a way is connected to exactly two edges.

        :return: List of connections.
        """
        connections = []
        n_connections = 0
        n_crossings = len(self.crossings)
        for i_way, way in enumerate(self.ways):
            # Check for starting node and end node of way if it is connected to two vertices.
            # That would mean that it is attached to another way. Then we check if it is not
            # already processed.
            if (
                self.parms.n_edges_per_vertex[self.parms.ivs.index(way.ivs[0])] == 2  # noqa: PLR2004
                and way.parms.connection.i_start == -1
            ):
                # We found a connection, but no connection is constructed. So construct connection
                # with other way.
                for j, way2 in enumerate(self.ways):
                    if i_way != j and (way2.ivs[0] == way.ivs[0] or way2.ivs[-1] == way.ivs[0]):
                        connections.append(Connection(n_connections, way, way2))
                        if connections[-1].crossing is not None:
                            connections[-1].crossing.idx = n_crossings
                            n_crossings += 1
                        n_connections += 1
                        break
            if (
                self.parms.n_edges_per_vertex[self.parms.ivs.index(way.ivs[-1])] == 2  # noqa: PLR2004
                and way.parms.connection.i_end == -1
            ):
                # We found a connection, but no connection is constructed. So construct connection
                # with other way
                for j, way2 in enumerate(self.ways):
                    if i_way != j and (way2.ivs[0] == way.ivs[-1] or way2.ivs[-1] == way.ivs[-1]):
                        connections.append(Connection(n_connections, way, way2))
                        if connections[-1].crossing is not None:
                            connections[-1].crossing.idx = n_crossings
                            n_crossings += 1
                        n_connections += 1
                        break
        return connections

    def process(self) -> None:
        """Process the road network to make it ready for plotting."""
        # Process all crossings
        for crossing in self.crossings:
            crossing.process()

        # Process all connections
        for i, connection in enumerate(self.connections):
            new_vertices, i_crossings = connection.process()
            for vertex in new_vertices:
                vertex.idx = np.max(self.parms.ivs) + 1
                self.vertices.append(vertex)
                self.parms.ivs.append(vertex.idx)
            for crossing in self.crossings:
                if crossing.idx in i_crossings:
                    crossing.process()
            for other_connection in self.connections[:i]:
                if (
                    other_connection.crossing is not None
                    and other_connection.crossing.idx in i_crossings
                ):
                    other_connection.process()

        # Process all ways
        for way in self.ways:
            way.process()

    def plot(self) -> Tuple[Figure, Axes]:
        """Plot the road network.

        :return: plot handle.
        """
        self.parms.axes.set_facecolor(self.options.face_color)

        # Plot the crossings
        for crossing in self.crossings:
            if not crossing.parms.part_of_big_crossing:
                crossing.plot(self.parms.axes)

        # Plot connection if it is a crossing
        for connection in self.connections:
            if connection.crossing is not None:
                connection.crossing.plot(self.parms.axes)

        # Plots the roads
        for way in self.ways:
            x_plot = np.array([way.parms.left.x_data, np.flipud(way.parms.right.x_data)])
            y_plot = np.array([way.parms.left.y_data, np.flipud(way.parms.right.y_data)])
            self.parms.axes.add_patch(
                patches.Polygon(
                    np.concatenate(
                        (
                            np.reshape(x_plot, (np.prod(x_plot.shape), 1)),
                            np.reshape(y_plot, (np.prod(y_plot.shape), 1)),
                        ),
                        axis=1,
                    ),
                    color=way.options.color,
                    zorder=way.options.layer,
                )
            )
            if way.options.show_border:
                self.parms.axes.plot(
                    x_plot.T, y_plot.T, color=way.options.side_color, zorder=way.options.layer
                )
            if way.parms.plot.lines != []:
                lines = np.array(way.parms.plot.lines)
                self.parms.axes.plot(
                    [lines[:, 0], lines[:, 2]],
                    [lines[:, 1], lines[:, 3]],
                    zorder=way.options.layer,
                    color=way.options.line_color,
                )
            way.plot_markers(self.parms.axes)

        return self.parms.figure, self.parms.axes

    def find_crossing(self, way: Way) -> Union[CrossingInfo, None]:
        """Find a crossing that is connected to the given way.

        :param way: The way for which a crossing has to be found.
        :return: Information about the crossing (if found), otherwise None.
        """
        if way.parms.crossing.i_start >= 0:
            crossing = next(
                crossing
                for crossing in self.crossings
                if crossing.idx == way.parms.crossing.i_start
            )
            lam = max(way.parms.crossing.start_lambda_left, way.parms.crossing.start_lambda_right)
            return CrossingInfo(crossing=crossing, lam=lam, crossing_at_start=True)
        if way.parms.crossing.i_end >= 0:
            crossing = next(
                crossing for crossing in self.crossings if crossing.idx == way.parms.crossing.i_end
            )
            lam = max(way.parms.crossing.end_lambda_left, way.parms.crossing.end_lambda_right)
            return CrossingInfo(crossing=crossing, lam=lam, crossing_at_start=False)
        return None

    def add_stopline(
        self,
        way: Way,
        info: Optional[CrossingInfo] = None,
        stoplineoptions: Optional[StopLineOptions] = None,
    ) -> bool:
        """Add a stopping line to the given way.

        :param way: The way that needs a stopping line.
        :param info: Optional information about the crossing.
        :param stoplineoptions: Options for the optional stop sign.
        :return: Whether the stopline has been added.
        """
        if stoplineoptions is None:
            stoplineoptions = StopLineOptions()
        if not stoplineoptions.stopline:
            return False
        if info is None:
            info = self.find_crossing(way)
            if info is None:
                return False
        if info.crossing_at_start and way.options.oneway:
            return False

        offset = way.options.nlanes * way.options.lanewidth / 2
        leftoffset = -offset
        if way.options.oneway:
            rightoffset = offset
        else:
            rightoffset = (way.options.nlanes / 2 - way.options.nlanes // 2) * way.options.lanewidth
        if self.options.rightdrive and not way.options.oneway:
            leftoffset, rightoffset = -rightoffset, -leftoffset
        left = info.crossing.start_and_direction(way, leftoffset, left=True)
        right = info.crossing.start_and_direction(way, rightoffset, left=True)
        lamoffset = stoplineoptions.lonoffset / np.hypot(left.x_dir, left.y_dir)
        self.parms.axes.plot(
            [
                left.x_init + (info.lam + lamoffset) * left.x_dir,
                right.x_init + (info.lam + lamoffset) * right.x_dir,
            ],
            [
                left.y_init + (info.lam + lamoffset) * left.y_dir,
                right.y_init + (info.lam + lamoffset) * right.y_dir,
            ],
            color=way.options.line_color,
            zorder=way.options.layer,
        )

        if stoplineoptions.stopsign:
            self.add_stopsign(way, info, lonoffset=stoplineoptions.lonoffset)
        self.add_dirsigns(way, info, stoplineoptions.dir_signs)
        return True

    def add_stopsign(self, way: Way, info: CrossingInfo, lonoffset: float = 0) -> None:
        """Add a stop SIGN to a given way.

        :param way: The way that needs a stopping line.
        :param info: Information about the crossing.
        :param lonoffset: Further offset of the line from the crossing.
        """
        for ilane in range(way.options.nlanes):
            if (
                way.options.oneway
                or (ilane < way.options.nlanes / 2 and not self.options.rightdrive)
                or (ilane >= way.options.nlanes / 2 and self.options.rightdrive)
            ):
                position = info.crossing.start_and_direction(
                    way, way.options.lanewidth * (ilane - way.options.nlanes / 2 + 0.5), left=True
                )
                distance = np.hypot(position.x_dir, position.y_dir)
                lamoffset = lonoffset / distance
                width = way.options.lanewidth - 0.5
                options = LettersOptions(
                    width=width,
                    length=width,
                    face_color=self.options.marker_fill_color,
                    edge_color=self.options.marker_border_color,
                )
                road_sign = Letters(self.parms.axes, "STOP", options=options)
                road_sign.change_pos(
                    position.x_init
                    + (info.lam + lamoffset) * position.x_dir
                    + (width / 2 + 0.5) * position.x_dir / distance,
                    position.y_init
                    + (info.lam + lamoffset) * position.y_dir
                    + (width / 2 + 0.5) * position.y_dir / distance,
                    np.arctan2(-position.x_dir, -position.y_dir),
                )

    def add_dirsigns(
        self, way: Way, info: CrossingInfo, dir_signs: Union[List[Union[str, None]], None]
    ) -> None:
        """Add directional signs to a given way.

        :param way: The way that needs a stopping line.
        :param info: Information about the crossing.
        :param dir_signs: The directional signs.
        """
        if dir_signs is not None:
            for ilane, dir_sign in enumerate(dir_signs):
                if dir_sign:
                    position = info.crossing.start_and_direction(
                        way,
                        way.options.lanewidth * (way.options.nlanes / 2 - ilane - 0.5),
                        left=True,
                    )
                    distance = np.hypot(position.x_dir, position.y_dir)
                    width = way.options.lanewidth - 0.5
                    options = TurnArrowOptions(
                        face_color=self.options.marker_fill_color,
                        edge_color=self.options.marker_border_color,
                    )
                    turn_arrow = TurnArrow(self.parms.axes, direction=dir_sign, options=options)
                    turn_arrow.change_pos(
                        position.x_init
                        + info.lam * position.x_dir
                        + (width + 2.25) * position.x_dir / distance,
                        position.y_init
                        + info.lam * position.y_dir
                        + (width + 2.25) * position.y_dir / distance,
                        np.arctan2(-position.x_dir, -position.y_dir),
                    )
                    self.parms.turn_arrows.append(turn_arrow)

    def add_traffic_lights(
        self,
        crossing: Optional[Crossing] = None,
        leftright: Optional[Tuple[bool, bool]] = None,
        options: Optional[TrafficLightOptions] = None,
        stoplineoptions: Optional[StopLineOptions] = None,
    ) -> List[TrafficLight]:
        """Add the traffic lights. It is only useful to call this function after plot().

        It will add traffic lights on both sides of a way, if a way is connected to a
        crossing and the way contains lane information through the turnlanes field.

        :param crossing: Crossing for adding the traffic lights. If none, all crossings.
        :param leftright: Whether to plot light on the left/right side (default=(True, True)).
        :param options: Options for the traffic light.
        :param stoplineoptions: Options for stopline. If none provided, no stopline is added.
        :return: List of traffic lights that have been added.
        """
        traffic_lights: List[TrafficLight] = []
        if leftright is None or (leftright[0] and leftright[1]):
            location = [False, True]
        elif leftright[0]:
            location = [False]
        elif leftright[1]:
            location = [True]
        else:
            return traffic_lights
        ways = self.ways if crossing is None else crossing.ways
        for way in ways:
            info = self.find_crossing(way)
            if info is None:
                continue

            if info.crossing_at_start and way.options.oneway:
                continue

            offset = way.options.nlanes * way.options.lanewidth / 2 + 1
            for left in location:
                orientation = info.crossing.start_and_direction(way, offset, left=left)
                traffic_light = TrafficLight(self.parms.axes, options=options)
                traffic_lights.append(traffic_light)
                self.parms.traffic_lights.append(traffic_light)
                self.parms.traffic_lights[-1].change_pos(
                    orientation.x_init + info.lam * orientation.x_dir,
                    orientation.y_init + info.lam * orientation.y_dir,
                    np.arctan2(-orientation.x_dir, -orientation.y_dir),
                )

            if stoplineoptions is not None:
                self.add_stopline(way, info=info, stoplineoptions=stoplineoptions)
        return traffic_lights



"""Ambulance object.

Author(s): Erwin de Gelder
"""

from typing import Optional, Tuple

import numpy as np
from matplotlib.axes import Axes

from .bus import Bus, BusOptions
from .polygon import Polygon


class AmbulanceOptions(BusOptions):
    """The default values of the options of an ambulance.

    The following list shows the options (within parentheses default values):
        length (4.8): The length of the ambulance.
        width (2.5): The width of the ambulance.
        line_width (1): The line width for plotting the ambulance.
        x_position_init (0): The initial horizontal position of the ambulance.
        y_position_init (0): The initial vertical position of the ambulance.
        angle_init (0): The initial angle of the ambulance.
        color ((.8, .8, .8)): Main color of the ambulance.
        color2 (None): Secondary color of the truck. By default, lighter version
            of color.
        luminance_diff (0.3): If color2 is not set, specify luminance difference
            between color2 and color.
        window_color (.2, .2, .2): Color of the windscreen.
        front_light_color (1, 1, 1): Color of front light beam.
        emergency_color (.9, .15, .15): Color of emergency light.
        hospital_sign_color (.25, .5, .75): Color of the hospital signs.
        edgecolor ((0, 0, 0)): Not used.
        layer (2): The layer in which the ambulance will be plotted.
    """

    length: float = 4.8
    width: float = 2.5
    color: Tuple[float, float, float] = (0.8, 0.8, 0.8)
    color2: Optional[Tuple[float, float, float]] = None
    luminance_diff: float = 0.3
    window_color: Tuple[float, float, float] = (0.2, 0.2, 0.2)
    front_light_color: Tuple[float, float, float] = (1, 1, 1)
    emergency_color: Tuple[float, float, float] = (0.9, 0.15, 0.15)
    hospital_sign_color: Tuple[float, float, float] = (0.25, 0.5, 0.75)


class Ambulance(Bus):
    """Plot an ambulance on the scene.

    Create an ambulance object using Ambulance(axes, options).

    Attributes:
        options (TruckOptions): All options. For a detailed description, see above.
        axes (Axes): The axes that is used for plotting.
    """

    def __init__(self, axes: Axes, options: Optional[AmbulanceOptions] = None) -> None:
        """Create an ambulance.

        :param axes: Axes on which the ambulance has to be plotted.
        :param options: Options for configuring the appearance of the ambulance.
        """
        if options is None:
            options = AmbulanceOptions()
        Bus.__init__(self, axes, options)
        self.options: AmbulanceOptions

    def plot_vehicle(self) -> None:
        """Plot the ambulance."""
        yoffset = 158
        yscale = 304
        xoffset = 81.5
        xscale = 158

        # Fill with main color, mirror.
        xdata = np.array([37, 30, 24, 21, 20, 20, 12, 9, 6, 5, 5, 9, 15, 20, 20, 21, 25, 28])
        ydata = np.array(
            [6, 8, 13, 18, 23, 96, 96, 97, 100, 103, 105, 105, 104, 104, 302, 305, 309, 310]
        )
        xdata = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                -ydata * self.options.length,
                facecolor=self.options.color,
                edgecolor=None,
                zorder=self.options.layer,
            ),
        )

        # Secondary, mirror
        xdatas = (
            np.array([36, 33, 27, 26, 26, 28, 47, 52, 52, 55, 68]),
            np.array([65, 56, 50, 45, 42, 40, 40, 42]),
        )
        ydatas = (
            np.array([16, 17, 23, 26, 88, 88, 83, 29, 83, 82, 81]),
            np.array([118, 119, 120, 121, 122, 124, 297, 299]),
        )
        for xdata, ydata in zip(xdatas, ydatas):
            xdata_new = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
            ydata_new = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
            self.fills += (
                Polygon(
                    self.axes,
                    xdata_new * self.options.width,
                    -ydata_new * self.options.length,
                    facecolor=Bus.determine_color2(self),
                    edgecolor=None,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Front light beam, 2x
        xdata = np.array([54, 36, 32, 27, 27, 29, 34, 39, 54])
        ydata = np.array([8, 8, 9, 13, 19, 17, 14, 13, 13])
        for xdata2 in (xdata, 2 * xoffset - xdata):
            self.fills += (
                Polygon(
                    self.axes,
                    (xdata2 - xoffset) / xscale * self.options.width,
                    -(ydata - yoffset) / yscale * self.options.length,
                    facecolor=self.options.front_light_color,
                    edgecolor=None,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Emergency lights, 2x
        for xdata, ydata in (
            (np.array([70, 45, 45, 70]), np.array([145, 145, 167, 167])),
            (np.array([32, 25, 25, 27, 32]), np.array([168, 168, 300, 300, 298])),
            (np.array([54, 32, 32, 54]), np.array([305, 305, 308, 308])),
        ):
            for xdata2 in (xdata, 2 * xoffset - xdata):
                self.fills += (
                    Polygon(
                        self.axes,
                        (xdata2 - xoffset) / xscale * self.options.width,
                        -(ydata - yoffset) / yscale * self.options.length,
                        facecolor=self.options.emergency_color,
                        edgecolor=None,
                        zorder=self.options.layer,
                        fixed_color=True,
                    ),
                )

        # Main
        xdata = (np.array([70, 70, 93, 93]) - xoffset) / xscale
        ydata = (np.array([145, 167, 167, 145]) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                -ydata * self.options.length,
                facecolor=self.options.color,
                edgecolor=None,
                zorder=self.options.layer,
            ),
        )

        # Hospital signs
        for ysign, rsign, dsign in ((47, 16, 6), (255, 28, 10.5)):
            xdata = np.zeros(18)
            ydata = np.zeros(18)
            theta = 0.0
            for i in range(6):
                xdata[i * 3] = np.sin(theta) * rsign + np.cos(theta) * dsign / 2
                ydata[i * 3] = np.cos(theta) * rsign - np.sin(theta) * dsign / 2
                xdata[i * 3 + 1] = (
                    np.sin(theta) * dsign / 2 * np.sqrt(3) + np.cos(theta) * dsign / 2
                )
                ydata[i * 3 + 1] = (
                    np.cos(theta) * dsign / 2 * np.sqrt(3) - np.sin(theta) * dsign / 2
                )
                theta += np.pi / 3
                xdata[i * 3 + 2] = np.sin(theta) * rsign - np.cos(theta) * dsign / 2
                ydata[i * 3 + 2] = np.cos(theta) * rsign + np.sin(theta) * dsign / 2
            self.fills += (
                Polygon(
                    self.axes,
                    xdata / xscale * self.options.width,
                    (yoffset - ysign) / yscale * self.options.length
                    - ydata / xscale * self.options.width,
                    facecolor=self.options.hospital_sign_color,
                    edgecolor=None,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )

        # Windscreen, mirror
        xdata = np.array([68, 55, 52, 47, 28, 42, 45, 50, 56, 65])
        ydata = np.array([81, 82, 83, 83, 88, 122, 121, 120, 119, 118])
        xdata = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                -ydata * self.options.length,
                facecolor=self.options.window_color,
                edgecolor=None,
                zorder=self.options.layer,
                fixed_color=True,
            ),
        )

        # Windscreen, 2x
        xdata = np.array([25, 38, 38, 25])
        ydata = np.array([168, 168, 127, 97])
        for xdata2 in (xdata, 2 * xoffset - xdata):
            self.fills += (
                Polygon(
                    self.axes,
                    (xdata2 - xoffset) / xscale * self.options.width,
                    -(ydata - yoffset) / yscale * self.options.length,
                    facecolor=self.options.window_color,
                    edgecolor=None,
                    zorder=self.options.layer,
                    fixed_color=True,
                ),
            )



"""Scripts for testing static_objects.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from traffic_scene_renderer import (
    Building,
    MaxSpeed,
    MaxSpeedOptions,
    Stripes,
    StripesOptions,
    TurnArrow,
    TurnArrowOptions,
)

mpl.use("Agg")


def save_fig(figure: Figure, axes: Axes, filename: Path, fwidth: float = 10) -> None:
    filename = Path("tests") / "created_images" / filename
    Path.unlink(filename, missing_ok=True)
    Path.mkdir(filename.parent, parents=True, exist_ok=True)
    xlim = axes.get_xlim()
    ylim = axes.get_ylim()
    figure.set_size_inches((fwidth, fwidth * (ylim[1] - ylim[0]) / (xlim[1] - xlim[0])))
    axes.set_axis_off()
    figure.savefig(filename, bbox_inches="tight", pad_inches=0.0, facecolor=axes.get_facecolor())
    plt.close(figure)


def test_max_speed_creation() -> None:
    fig, axes = plt.subplots()
    MaxSpeed(axes, "50", MaxSpeedOptions(fontsize=50))
    save_fig(fig, axes, Path("static_objects") / "max_speed.png", 2)


def test_max_speed_different_radii() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-4.5, 4.5)
    axes.set_ylim(-2, 2)
    MaxSpeed(axes, "50", MaxSpeedOptions(fontsize=30, outer_radius=1.2, inner_radius=0.8))
    max_speed2 = MaxSpeed(axes, "50", MaxSpeedOptions(fontsize=45, outer_radius=1.5))
    max_speed2.change_pos(-3, 0)
    max_speed3 = MaxSpeed(axes, "50", MaxSpeedOptions(fontsize=20, inner_radius=0.53))
    max_speed3.change_pos(3, 0)
    save_fig(fig, axes, Path("static_objects") / "max_speed_different_sizes.png", 6)


def test_creation_of_all_turn_arrows() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-0.5, 16.5)
    axes.set_ylim(-1.5, 4.5)
    for i, direction in enumerate(
        ("through", "left", "right", "leftright", "leftthrough", "throughright")
    ):
        turn_arrow = TurnArrow(
            axes, direction=direction, options=TurnArrowOptions(face_color=(1 - 0.1 * i, 0.8, 0.8))
        )
        turn_arrow.change_pos(i * 3, 0)
        turn_arrow = TurnArrow(
            axes, direction=direction, options=TurnArrowOptions(face_color=(0.8, 1 - 0.1 * i, 0.8))
        )
        turn_arrow.change_pos(i * 3, 3, np.pi)
    save_fig(fig, axes, Path("static_objects") / "turn_arrows.png", 10)


def test_building_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-2, 2)
    Building(axes)
    save_fig(fig, axes, Path("static_objects") / "building.png", 2)


def test_stripes_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-2, 2)
    Stripes(axes, options=StripesOptions(width=0.5, length=3.5, nstripes=20))
    save_fig(fig, axes, Path("static_objects") / "stripes.png", 3)


def test_stripes_change_colors() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-2, 2)
    stripes = Stripes(axes, options=StripesOptions(width=0.5, length=3.5))
    stripes.change_color(face_color=(0, 0, 0), edge_color=(1, 0, 0), face_color2=(1, 1, 0))
    save_fig(fig, axes, Path("static_objects") / "stripes_different_color.png", 3)


def test_stripes_rotate() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-2, 2)
    stripes = Stripes(axes, options=StripesOptions(width=0.5, length=3.5))
    stripes.change_pos(0, 0, np.pi / 2)
    save_fig(fig, axes, Path("static_objects") / "stripes_rotated.png", 3)



"""Scripts for testing crossing.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.axes import Axes

from traffic_scene_renderer import Crossing, CrossingOptions, Vertex, Way, WayOptions
from traffic_scene_renderer.crossing import correct_angle

from .test_static_objects import save_fig
from .test_way import plot_way

mpl.use("Agg")


def plot_crossing(axes: Axes, *crossings: Crossing) -> None:
    ways_plotted = set()
    for crossing in crossings:
        crossing.process()
        crossing.plot(axes)
    for crossing in crossings:
        for way in crossing.ways:
            if way in ways_plotted:
                continue
            ways_plotted.add(way)
            way.process()
            plot_way(axes, way)


def crossing_creation(*, have_vertex_0_first: bool) -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-10, 10)
    vertices = [
        Vertex(*ixy) for ixy in ((0, 0, 0), (1, -10, 0), (2, 0, 10), (3, 10, 0), (4, 0, -10))
    ]
    if have_vertex_0_first:
        ways = [
            Way([vertices[0], vertices[i]], WayOptions(line_color=(0, 0, 0)))
            for i in range(1, len(vertices))
        ]
    else:
        ways = [
            Way([vertices[i], vertices[0]], WayOptions(line_color=(0, 0, 0)))
            for i in range(1, len(vertices))
        ]
    crossing = Crossing(0, vertices[0], ways)
    plot_crossing(axes, crossing)
    save_fig(
        fig, axes, Path("crossing") / f"simple_crossing{1 if have_vertex_0_first else 2}.png", 10
    )


def test_crossing_creation() -> None:
    crossing_creation(have_vertex_0_first=True)


def test_crossing_creation_reversed() -> None:
    crossing_creation(have_vertex_0_first=False)


def crossing_large_angle(filename: str, *, roundabout: bool) -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-10, 10)
    vertices = [Vertex(*ixy) for ixy in ((0, 0, 0), (1, -10, 1), (2, 0, 10), (3, 10, 1))]
    ways = [
        Way(way_vertices, WayOptions(line_color=(0, 0, 0)))
        for way_vertices in (
            [vertices[0], vertices[1]],
            [vertices[0], vertices[2]],
            [vertices[3], vertices[0]],
        )
    ]
    crossing = Crossing(0, vertices[0], ways, CrossingOptions(roundabout=roundabout))
    plot_crossing(axes, crossing)
    save_fig(fig, axes, Path("crossing") / filename, 10)


def test_crossing_large_angle() -> None:
    crossing_large_angle("crossing_large_angle.png", roundabout=False)


def test_roundabout() -> None:
    crossing_large_angle("roundabout.png", roundabout=True)


def create_zebra(*, with_blocks: bool) -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-7, 7)
    vertices = [
        Vertex(*ixy) for ixy in ((0, 0, 0), (1, -10, 0), (2, 0, 10), (3, 10, 0), (4, 0, -10))
    ]
    ways = [
        Way([vertices[0], vertices[1]], WayOptions(line_color=(0, 0, 0))),
        Way([vertices[0], vertices[2]], WayOptions(side_color=(0, 0, 0), highway="footway")),
        Way([vertices[0], vertices[3]], WayOptions(line_color=(0, 0, 0))),
        Way([vertices[0], vertices[4]], WayOptions(side_color=(0, 0, 0), highway="footway")),
    ]
    ways[1].options.highway = "footway"
    ways[3].options.highway = "footway"
    crossing = Crossing(0, vertices[0], ways, CrossingOptions(zebra_square_markings=with_blocks))
    plot_crossing(axes, crossing)
    save_fig(
        fig,
        axes,
        Path("crossing") / f"zebra_crossing{'_with_blocks' if with_blocks else ''}.png",
        10,
    )


def test_zebra() -> None:
    create_zebra(with_blocks=False)


def test_zebra_with_blocks() -> None:
    create_zebra(with_blocks=True)


def test_footway_different_radii() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-10, 10)
    axes.set_ylim(-7, 7)
    vertices = [
        Vertex(*ixy)
        for ixy in (
            (0, -10, 0),
            (1, -3, 0),
            (2, -3, 10),
            (3, -3, -10),
            (4, 4, 1),
            (5, 2, 10),
            (6, 2, -10),
            (7, 12, 8),
            (8, 12, -8),
        )
    ]
    ways = [
        Way([vertices[i], vertices[j]], WayOptions(side_color=(1, 0, 0), highway="footway"))
        for i, j in ((0, 1), (1, 2), (1, 3), (1, 4), (4, 5), (4, 6), (4, 7), (4, 8))
    ]
    crossing1 = Crossing(0, vertices[1], ways[:4])
    crossing2 = Crossing(1, vertices[4], ways[3:], CrossingOptions(radius=1))
    plot_crossing(axes, crossing1, crossing2)
    save_fig(fig, axes, Path("crossing") / "footways.png", 10)


def test_warning_zebra_no_2_footways() -> None:
    fig, axes = plt.subplots()
    vertices = [Vertex(0, 0, 0), Vertex(1, 10, 0), Vertex(2, 0, 10), Vertex(3, 0, -10)]
    ways = [Way(vertices[:2]), Way(vertices[1:3]), Way(vertices[2:])]
    crossing = Crossing(0, vertices[0], ways)
    crossing.parms.zebra = True
    with pytest.warns(UserWarning):
        crossing.plot_zebra(axes)
    plt.close(fig)


def test_warning_zebra_no_2_roadways() -> None:
    fig, axes = plt.subplots()
    vertices = [Vertex(0, 0, 0), Vertex(1, 10, 0), Vertex(2, 0, 10), Vertex(3, 0, -10)]
    ways = [
        Way(vertices[:2], WayOptions(highway="footway")),
        Way(vertices[1:3], WayOptions(highway="footway")),
        Way(vertices[2:]),
    ]
    crossing = Crossing(0, vertices[0], ways)
    crossing.parms.zebra = True
    with pytest.warns(UserWarning):
        crossing.plot_zebra(axes)
    plt.close(fig)


def test_correct_angle() -> None:
    assert correct_angle(0, 1.5 * np.pi, big_angle=False) == pytest.approx(2 * np.pi)
    assert correct_angle(1.5 * np.pi, 0, big_angle=False) == pytest.approx(-0.5 * np.pi)
    assert correct_angle(0, 0.5 * np.pi, big_angle=True) == pytest.approx(2 * np.pi)
    assert correct_angle(0.5 * np.pi, 0, big_angle=True) == pytest.approx(-1.5 * np.pi)



"""Car object.

Author(s): Erwin de Gelder
"""

from enum import Enum
from typing import List, Optional, Tuple

import numpy as np
from matplotlib.axes import Axes

from .path_follower import PathFollower
from .polygon import Polygon
from .vehicle import Vehicle, VehicleOptions


class CarType(Enum):
    """Different car types for drawing a transparent car."""

    VEHICLE = 1
    STATION_WAGON = 2
    PICKUP_TRUCK = 3
    SEDAN = 4


class InvalidCarTypeError(Exception):
    """Error to be raised in case a car type is specified that is not implemented."""

    def __init__(self, car_type: int) -> None:
        """Description of error."""
        super().__init__(
            f"Car type '{car_type}' is invalid. Choose one of the CarType enumeration."
        )


class CarOptions(VehicleOptions):
    """The default values of the options of a car.

    The following list shows the options with within parentheses the default values:
        length (4.5): The length of the car.
        width (1.8): The width of the car.
        icar (CarType.VEHICLE): The index of the type of the car.
        line_width (1): The line width for plotting the car.
        x_position_init (0): The initial horizontal position of the car.
        y_position_init (0): The initial vertical position of the car.
        angle_init (0): The initial angle of the car.
        fill (True): Whether to use a non-transparent car (True) or a transparent car (False).
        color ([0, 0.4375, 0.75]): Color of the car.
        windowcolor ([1, 1, 1]): In case of a non-transparent car, the color of the windows can be
                                 set.
        front_light_color ([1, 1, 1]): Colors of the front light beam.
        edgecolor ([0, 0, 0]): In case of a non-transparent car, the color of the lines is set
                               using this option.
        layer (2): The layer in which the car will be plotted.
    """

    length: float = 4.5
    width: float = 1.8
    fill: bool = True
    icar: CarType = CarType.VEHICLE
    window_color: Tuple[float, float, float] = (1, 1, 1)
    front_light_color: Tuple[float, float, float] = (1, 1, 1)


class Car(Vehicle):
    """Plot a car on the scene.

    Attributes:
        options (CarOptions): All options. For a detailed description, see above.
        is_braking (bool): Whether the car is braking or not.
        axes (Axes): The axes that is used for plotting.
    """

    def __init__(
        self,
        axes: Axes,
        options: Optional[CarOptions] = None,
        path_follower: Optional[PathFollower] = None,
    ) -> None:
        """Create a car object.

        :param axes: The axes on which the car is supposed to be drawn.
        :param options: Various options for the car, see CarOptions.
        :param path_follower: Path that the car needs to follow (if any).
        """
        if options is None:
            options = CarOptions()
        Vehicle.__init__(self, axes, options, path_follower)
        self.options: CarOptions

    def plot_vehicle(self) -> None:
        """Do the actual plotting of the vehicle on the specified axes."""
        if self.options.fill:
            self.draw_filled_car()
        else:
            self.draw_transparent_car()

    def car(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Get (x,y) data of the lines for drawing the (transparent) car.

        :return: The (x,y) data of the lines for drawing the car.
        """
        if self.options.icar == CarType.VEHICLE:
            xdata = [
                [2.0, 2, 5, 6, 6, 5, 2, 2, 6, 13, 13, 13, 16, 61, 63, 63],
                [61.0, 67, 76, 80, 80],
                [67.0, 14, 6, 14, 67, 64, 67, 95, 101, 107, 105, 99, 101, 107, 109, 111, 111],
            ]
            ydata = [
                [26.0, 14, 14, 15, 20, 21, 21, 13, 7, 11, 26, 11, 9, 9, 13, 26],
                [9.0, 4, 12, 20, 26],
                [4.0, 2, 7, 2, 4, 1, 4, 3, 5, 12, 13, 9, 5, 12, 16, 23, 26],
            ]
            xoffset = 56.5
            xscale = 109.0
            yoffset = 26.0
            yscale = 50.0
        elif self.options.icar == CarType.STATION_WAGON:
            xdata = [
                [49.0, 49, 51, 109, 112, 113, 113],
                [109.0, 114, 120, 125, 128, 128],
                [114.0, 81, 81, 81, 52, 51, 52, 55, 139, 146, 142, 138, 138],
                [146.0, 148, 148],
            ]
            ydata = [
                [68.0, 60, 53, 53, 60, 63, 68],
                [53.0, 49, 49, 54, 62, 68],
                [49.0, 50, 53, 50, 50, 53, 50, 47, 47, 54, 54, 50, 47],
                [54.0, 61, 68],
            ]
            xoffset = 98.5
            xscale = 99
            yoffset = 68
            yscale = 42
        elif self.options.icar == CarType.PICKUP_TRUCK:
            xdata = [
                [29.0, 29, 78, 78],
                [29.0, 78, 78, 29, 29],
                [29.0, 78, 78, 29, 29],
                [11.0, 11, 14, 19, 19, 19, 22, 135, 129, 135, 164, 177, 183, 183],
                [84.0, 84, 93, 93],
                [129.0, 129, 126, 146, 149, 149],
                [25.0, 25, 79],
            ]
            ydata = [
                [49.5, 49, 49, 49.5],
                [39.0, 39, 40, 40, 39],
                [28.0, 28, 29, 29, 28],
                [49.5, 22, 19, 19, 49.5, 16, 13, 14, 8, 14, 15, 20, 45, 49.5],
                [49.5, 19, 23, 49.5],
                [49.5, 37, 22, 18, 38, 49.5],
                [49.5, 20, 20],
            ]
            xoffset = 97
            xscale = 172
            yoffset = 49.5
            yscale = 73
        elif self.options.icar == CarType.SEDAN:
            xdata = [
                [1.0, 1, 5, 27, 100, 95, 100, 135, 149, 154, 154, 154, 121],
                [38.0, 51, 81, 95],
                [38.0, 38, 42, 30, 22, 18, 15, 15],
                [88.0, 88, 87, 111, 116, 116],
            ]
            ydata = [
                [39.0, 23, 12, 6, 6, 1, 6, 6, 12, 26, 38, 27, 14],
                [10.0, 14, 14, 10],
                [38.0, 31, 17, 17, 20, 24, 32, 38],
                [38.0, 25, 17, 10, 33, 38],
            ]
            xoffset = 77.5
            xscale = 153
            yoffset = 38
            yscale = 70  # Mirrors will be 'outside' car
        else:
            raise InvalidCarTypeError(self.options.icar)

        xdata_scaled = []
        ydata_scaled = []
        for xdata_sub, ydata_sub in zip(xdata, ydata):
            xdata_scaled.append((np.array(xdata_sub) - xoffset) / xscale * self.options.length)
            ydata_scaled.append((np.array(ydata_sub) - yoffset) / yscale * self.options.width)

        return xdata_scaled, ydata_scaled

    def draw_transparent_car(self) -> None:
        """Draw a transparent car."""
        xdata, ydata = self.car()
        for xdata_sub, ydata_sub in zip(xdata, ydata):
            self.plots += (
                self.axes.plot(
                    ydata_sub,
                    xdata_sub,
                    "r-",
                    zorder=self.options.layer,
                    linewidth=self.options.line_width,
                )[0],
            )
            self.plots += (
                self.axes.plot(
                    ydata_sub * -1,
                    xdata_sub,
                    "r-",
                    zorder=self.options.layer,
                    linewidth=self.options.line_width,
                )[0],
            )

    def draw_filled_car(self) -> None:
        """Draw a car that is not transparent."""
        yoffset = 278.5
        yscale = 555
        xoffset = 135.5
        xscale = 224

        # Fill blue, mirror.
        xdata = np.array([135, 70, 45, 36, 30, 23, 28, 35, 57, 81, 135])
        ydata = np.array([1, 2, 9, 19, 35, 92, 513, 524, 541, 548, 556])
        xdata_new = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
        ydata_new = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata_new * self.options.width,
                ydata_new * self.options.length,
                facecolor=self.options.color,
                edgecolor=self.options.edgecolor,
                zorder=self.options.layer,
            ),
        )

        # Fill blue, twice.
        xdata = (np.array([32, 7, 1, 4, 30]) - xoffset) / xscale
        ydata = (np.array([361, 360, 361, 365, 371]) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                ydata * self.options.length,
                facecolor=self.options.color,
                edgecolor=self.options.edgecolor,
                linewidth=self.options.line_width,
                zorder=self.options.layer,
            ),
            Polygon(
                self.axes,
                -xdata * self.options.width,
                ydata * self.options.length,
                facecolor=self.options.color,
                edgecolor=self.options.edgecolor,
                linewidth=self.options.line_width,
                zorder=self.options.layer,
            ),
        )

        # Black line, mirror.
        for xdata, ydata in zip(
            [np.array([43, 44, 51, 61]), np.array([82, 79, 96]), np.array([44, 40, 82])],
            [np.array([77, 49, 23, 15]), np.array([524, 527, 538]), np.array([493, 396, 524])],
        ):
            xdatas = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
            ydatas = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
            self.plots += (
                self.axes.plot(
                    xdatas * self.options.width,
                    ydatas * self.options.length,
                    color=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    zorder=self.options.layer,
                )[0],
            )

        # Fill white, mirror.
        for xdata, ydata in zip(
            [
                np.array([135, 89, 61, 46, 45, 62, 79, 105]),
                np.array([135, 81, 47, 40, 66, 90, 111]),
            ],
            [
                np.array([67, 69, 75, 83, 90, 153, 147, 146]),
                np.array([412, 409, 398, 391, 326, 332, 334]),
            ],
        ):
            xdatas = (np.concatenate((xdata, 2 * xoffset - np.flipud(xdata))) - xoffset) / xscale
            ydatas = (np.concatenate((ydata, np.flipud(ydata))) - yoffset) / yscale
            self.fills += (
                Polygon(
                    self.axes,
                    xdatas * self.options.width,
                    ydatas * self.options.length,
                    facecolor=self.options.window_color,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    fixed_color=True,
                    zorder=self.options.layer,
                ),
            )

        # Fill windows, twice.
        for xdata, ydata in zip(
            [
                np.array([32, 32, 51, 45]),
                np.array([53, 37, 33, 33, 38, 56, 60, 58]),
                np.array([63, 32, 32, 38, 60]),
                np.array([30, 31, 46, 58, 70, 91, 72]),
            ],
            [
                np.array([152, 178, 178, 165]),
                np.array([184, 184, 188, 253, 256, 250, 241, 208]),
                np.array([260, 276, 367, 366, 315]),
                np.array([494, 514, 529, 536, 538, 538, 527]),
            ],
        ):
            xdatas = (np.array(xdata) - xoffset) / xscale
            ydatas = (np.array(ydata) - yoffset) / yscale
            self.fills += (
                Polygon(
                    self.axes,
                    xdatas * self.options.width,
                    ydatas * self.options.length,
                    facecolor=self.options.window_color,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    fixed_color=True,
                    zorder=self.options.layer,
                ),
                Polygon(
                    self.axes,
                    -xdatas * self.options.width,
                    ydatas * self.options.length,
                    facecolor=self.options.window_color,
                    edgecolor=self.options.edgecolor,
                    linewidth=self.options.line_width,
                    fixed_color=True,
                    zorder=self.options.layer,
                ),
            )

        # Fill front beam, twice.
        xdata = (np.array([30, 31, 46, 58, 70, 91, 72]) - xoffset) / xscale
        ydata = (np.array([494, 514, 529, 536, 538, 538, 527]) - yoffset) / yscale
        self.fills += (
            Polygon(
                self.axes,
                xdata * self.options.width,
                ydata * self.options.length,
                facecolor=self.options.front_light_color,
                edgecolor=self.options.edgecolor,
                linewidth=self.options.line_width,
                fixed_color=True,
                zorder=self.options.layer,
            ),
            Polygon(
                self.axes,
                -xdata * self.options.width,
                ydata * self.options.length,
                facecolor=self.options.front_light_color,
                edgecolor=self.options.edgecolor,
                linewidth=self.options.line_width,
                fixed_color=True,
                zorder=self.options.layer,
            ),
        )



"""Arrow to indicate direction of an object.

Author(s): Erwin de Gelder
"""

from typing import Any, List, Union

import numpy as np
from matplotlib.axes import Axes

SVD_TOL = 0.001


def arrow(
    axes: Axes,
    xdata: Union[List, np.ndarray],
    ydata: Union[List, np.ndarray],
    size: float = 0.75,
    **kwargs: Any,  # noqa: ANN401  # Ignore because otherwise mypy will complain.
) -> None:
    """Draw an arrow on the axes.

    If only 2 coordinates are provided, the arrow will be a straight line.
    If 4 coordinates are provided, 2 straight lines are drawn based on the first two
    coordinates and the last 2 coordinates. These straight lines are connected using
    an ellipsoid. Obviously, this only works if these lines can be connected using
    an ellipsoid.
    If 4 coordinates are provided with 2 straight parallel lines, they are connected
    using a sinusoid.
    If more coordinates are given, a line is simply drawn through these coordinates.

    :param axes: The axes that is used for plotting.
    :param xdata: The x values of the coordinates.
    :param ydata: The y values of the coordinates.
    :param size: The size of the arrow.
    :param kwargs: Additional plotting parameters.
    """
    # Draw the arrow
    if len(xdata) == 4:  # noqa: PLR2004
        # Find center of ellipse
        matrix = np.array(
            [[ydata[1] - ydata[0], ydata[2] - ydata[3]], [xdata[0] - xdata[1], xdata[3] - xdata[2]]]
        )
        if np.linalg.svd(matrix)[1][-1] > SVD_TOL:
            xy_ellipse = compute_ellipse(xdata, ydata)
            axes.plot(
                np.concatenate(([xdata[0]], xy_ellipse[0], [xdata[3]])),
                np.concatenate(([ydata[0]], xy_ellipse[1], [ydata[3]])),
                **kwargs,
            )
        else:  # Almost singular, so connect with sinusoid.
            xy_sinus = compute_sinus(xdata, ydata)
            axes.plot(
                np.concatenate(([xdata[0]], xy_sinus[0] + xdata[1], [xdata[3]])),
                np.concatenate(([ydata[0]], xy_sinus[1] + ydata[1], [ydata[3]])),
                **kwargs,
            )
    else:
        axes.plot(xdata, ydata, **kwargs)

    # Draw the tip of the arrow
    theta = np.arctan2(xdata[-1] - xdata[-2], ydata[-1] - ydata[-2])
    rotation_mat = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    xydata = np.array([[-size, 0, size], [-size, 0, -size]])
    xydata = np.dot(rotation_mat, xydata)
    axes.plot(xdata[-1] + xydata[0], ydata[-1] + xydata[1], **kwargs)


def compute_ellipse(xdata: Union[List, np.ndarray], ydata: Union[List, np.ndarray]) -> np.ndarray:
    """Compute an ellipse connecting the four points.

    :param xdata: The x values of the coordinates.
    :param ydata: The y values of the coordinates.
    :return: The (x,y) data of the ellipse.
    """
    matrix = np.array(
        [[ydata[1] - ydata[0], ydata[2] - ydata[3]], [xdata[0] - xdata[1], xdata[3] - xdata[2]]]
    )
    vector = np.array([xdata[2] - xdata[1], ydata[2] - ydata[1]])
    solution = np.linalg.solve(matrix, vector)
    center = np.array(
        [xdata[1] + matrix[0, 0] * solution[0], ydata[1] + matrix[1, 0] * solution[0]]
    )
    matrix_transform = np.array(
        [[xdata[2] - center[0], xdata[1] - center[0]], [ydata[2] - center[1], ydata[1] - center[1]]]
    )
    xy_transformed = np.array(
        [np.sin(np.linspace(0, np.pi / 2, 20)), np.cos(np.linspace(0, np.pi / 2, 20))]
    )
    xy_ellipse = np.dot(matrix_transform, xy_transformed)
    xy_ellipse[0] += center[0]
    xy_ellipse[1] += center[1]
    return xy_ellipse


def compute_sinus(xdata: Union[List, np.ndarray], ydata: Union[List, np.ndarray]) -> np.ndarray:
    """Compute a sinus connecting the four points.

    :param xdata: The x values of the coordinates.
    :param ydata: The y values of the coordinates.
    :return: The (x,y) data of the sinus.
    """
    matrix = np.array(
        [[xdata[1] - xdata[0], ydata[2] - ydata[3]], [ydata[1] - ydata[0], xdata[3] - xdata[2]]]
    )
    vector = np.array([xdata[2] - xdata[1], ydata[2] - ydata[1]])
    solution = np.linalg.solve(matrix, vector)
    lateral = np.cos(np.linspace(-np.pi, 0, 20)) / 2 + 0.5
    longitudinal = np.linspace(0, 1, 20)
    return np.array(
        [
            longitudinal * solution[0] * matrix[0, 0] + lateral * solution[1] * matrix[0, 1],
            longitudinal * solution[0] * matrix[1, 0] + lateral * solution[1] * matrix[1, 1],
        ]
    )



"""Constructing letters.

Author(s): Erwin de Gelder
"""

import sys
from typing import NamedTuple, Optional, Tuple

import numpy as np
from matplotlib.axes import Axes

from .static_objects import StaticObject, TurnArrowOptions

BIN_A = 65  # ord("A")
BIN_Z = 90  # ord("Z")
LetterOptions = TurnArrowOptions


class LetterData(NamedTuple):
    """Named tuple for storing letter data."""

    xdata: np.ndarray
    ydata: np.ndarray
    xcenter: float
    ycenter: float
    xdata_plot: Optional[Tuple[np.ndarray, ...]] = None
    ydata_plot: Optional[Tuple[np.ndarray, ...]] = None


class Letter(StaticObject):
    """Draw a letter (A to Z (capital) implemented)."""

    def __init__(self, axes: Axes, letter: str, options: Optional[LetterOptions] = None) -> None:
        """Create a single letter.

        :param axes: Axes on which the letter should be plotted.
        :param letter: The letter (A to Z; only capital supported).
        :param options: Any options for configuring the letter.
        """
        self.options = LetterOptions() if options is None else options
        StaticObject.__init__(self, axes)

        width, height = 3.5, 7
        if len(letter) == 1 and BIN_A <= ord(letter) <= BIN_Z:
            data = getattr(sys.modules[__name__], f"capital_{letter.lower()}")()  # type: LetterData
        else:
            msg = f"Letter '{letter:s}' is not yet implemented."
            raise NotImplementedError(msg)

        self.fills += (
            axes.fill(
                (data.xdata - data.xcenter) * self.options.width / width,
                (data.ydata - data.ycenter) * self.options.length / height,
                facecolor=self.options.face_color,
                edgecolor=None,
                zorder=self.options.layer,
            )[0],
        )
        if data.xdata_plot and data.ydata_plot:
            for xdata, ydata in zip(data.xdata_plot, data.ydata_plot):
                self.plots += (
                    axes.plot(
                        (np.concatenate((xdata, np.array([xdata[0]]))) - data.xcenter)
                        * self.options.width
                        / width,
                        (np.concatenate((ydata, np.array([ydata[0]]))) - data.ycenter)
                        * self.options.length
                        / height,
                        color=self.options.edge_color,
                        zorder=self.options.layer,
                    )[0],
                )
        else:
            self.plots += (
                axes.plot(
                    (np.concatenate((data.xdata, np.array([data.xdata[0]]))) - data.xcenter)
                    * self.options.width
                    / width,
                    (np.concatenate((data.ydata, np.array([data.ydata[0]]))) - data.ycenter)
                    * self.options.length
                    / height,
                    color=self.options.edge_color,
                    zorder=self.options.layer,
                )[0],
            )

    def change_color(
        self,
        face_color: Optional[Tuple[float, float, float]] = None,
        edge_color: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        """Change the colors of the letter.

        :param face_color: The new color of the filled area.
        :param edge_color: The new color of the edge of the filled areas and the lines.
        """
        StaticObject.change_color(self, face_color=face_color)
        if edge_color is not None:
            for plot in self.plots:
                plot.set_color(edge_color)


class LettersOptions(TurnArrowOptions):
    """Options for object with multiple letters."""

    margin_ratio: float = 0.4


class Letters(StaticObject):
    """A sign consisting of multiple letters."""

    def __init__(self, axes: Axes, sign: str, options: Optional[LettersOptions] = None) -> None:
        """Create a series of letters.

        :param axes: Axes on which the letters should be plotted.
        :param sign: String containing the letters.
        :param options: Any options for configuring the letter.
        """
        self.options = LettersOptions() if options is None else options
        StaticObject.__init__(self, axes)

        n_letters = len(sign)
        letter_width = self.options.width / (
            n_letters + (n_letters - 1) * self.options.margin_ratio
        )
        self.x_letters = (
            np.arange(n_letters) * (1 + self.options.margin_ratio) + 0.5
        ) * letter_width - self.options.width / 2
        letter_options = LetterOptions(
            width=letter_width,
            length=self.options.length,
            face_color=self.options.face_color,
            edge_color=self.options.edge_color,
            layer=self.options.layer,
        )
        self.letters = [Letter(axes, letter, options=letter_options) for letter in sign]
        for letter, xpos in zip(self.letters, self.x_letters):
            letter.change_pos(xpos, 0, 0)

    def change_color(
        self,
        face_color: Optional[Tuple[float, float, float]] = None,
        edge_color: Optional[Tuple[float, float, float]] = None,
    ) -> None:
        """Change the colors of the letters.

        :param face_color: The new face color of the letters (if any).
        :param edge_color: The new edge color of the letters (if any).
        """
        for letter in self.letters:
            letter.change_color(face_color, edge_color)

    def change_pos(self, xcenter: float, ycenter: float, angle: float = 0) -> None:
        """Change the position of the letters.

        :param xcenter: New x-coordinate.
        :param ycenter: New y-coordinate.
        :param angle: New angle.
        """
        for letter, xpos in zip(self.letters, self.x_letters):
            letter.change_pos(xcenter + np.cos(angle) * xpos, ycenter - np.sin(angle) * xpos, angle)


def capital_a() -> LetterData:
    """Return data for letter A."""
    xdata = np.array([1, 0, 0.875, 2.625, 3.5, 2.5, 2.25, 1.25, 1.375, 2.125, 1.75])
    ydata = np.array([0, 0, 7, 7, 0, 0, 2, 2, 3, 3, 6])
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=1.75,
        ycenter=3.5,
        xdata_plot=(xdata[:8], xdata[8:]),
        ydata_plot=(ydata[:8], ydata[8:]),
    )


def capital_b() -> LetterData:
    """Return data for letter B."""
    theta1 = np.linspace(0, np.arcsin(3 / 4) + np.pi / 2, 20)
    theta2 = np.linspace(0, np.pi, 20)
    xdata = np.concatenate(
        (
            [1, 0, 0, 1],
            1.5 + 2 * np.sin(theta1),
            1.5 + 2 * np.sin(np.flip(theta1)),
            [1, 1],
            1.5 + np.sin(theta2),
            [1, 1],
            1.5 + np.sin(theta2),
            [1],
        )
    )
    ydata = np.concatenate(
        (
            [0, 0, 7, 7],
            5 + 2 * np.cos(theta1),
            2 - 2 * np.cos(np.flip(theta1)),
            [0, 1],
            2 - np.cos(theta2),
            [3, 4],
            5 - np.cos(theta2),
            [6],
        )
    )
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=1.75,
        ycenter=3.5,
        xdata_plot=(xdata[:44], xdata[45:67], xdata[67:]),
        ydata_plot=(ydata[:44], ydata[45:67], ydata[67:]),
    )


def capital_c() -> LetterData:
    """Return data for letter C."""
    theta = np.linspace(0, np.pi, 20)
    return LetterData(
        xdata=np.concatenate(
            (
                [0.75, 1.75],
                1.75 * np.cos(theta),
                -1.75 * np.cos(theta),
                [1.75, 0.75],
                0.75 * np.cos(theta),
                -0.75 * np.cos(theta),
            )
        ),
        ydata=np.concatenate(
            (
                [5, 5],
                5.25 + 1.75 * np.sin(theta),
                1.75 - 1.75 * np.sin(theta),
                [2, 2],
                1.75 - 0.75 * np.sin(theta),
                5.25 + 0.75 * np.sin(theta),
            )
        ),
        xcenter=0,
        ycenter=3.5,
    )


def capital_d() -> LetterData:
    """Return data for letter D."""
    theta = np.linspace(0, np.pi / 2, 10)
    xdata = np.concatenate(
        (
            1.5 + 2 * np.sin(theta),
            1.5 + 2 * np.cos(theta),
            [0, 0, 1.5, 1.5, 1, 1],
            1.5 + np.sin(theta),
            1.5 + np.cos(theta),
        )
    )
    ydata = np.concatenate(
        (
            2 - 2 * np.cos(theta),
            5 + 2 * np.sin(theta),
            [7, 0, 0, 1, 1, 6],
            5 + np.cos(theta),
            2 - np.sin(theta),
        )
    )
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=1.75,
        ycenter=3.5,
        xdata_plot=(xdata[:22], xdata[23:]),
        ydata_plot=(ydata[:22], ydata[23:]),
    )


def capital_e() -> LetterData:
    """Return data for letter E."""
    return LetterData(
        xdata=np.array([0, 0, 3.5, 3.5, 1, 1, 3, 3, 1, 1, 3.5, 3.5]),
        ydata=np.array([7, 0, 0, 1, 1, 3, 3, 4, 4, 6, 6, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_f() -> LetterData:
    """Return data for letter F."""
    return LetterData(
        xdata=np.array([0, 3.5, 3.5, 1, 1, 2.5, 2.5, 1, 1, 0]),
        ydata=np.array([7, 7, 6, 6, 4, 4, 3, 3, 0, 0]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_g() -> LetterData:
    """Return data for letter G."""
    theta = np.linspace(0, np.pi, 20)
    return LetterData(
        xdata=np.concatenate(
            (
                [2.5, 3.5],
                1.75 + 1.75 * np.cos(theta),
                1.75 - 1.75 * np.cos(theta),
                [3.5, 1.5, 1.5, 2.5],
                1.75 + 0.75 * np.cos(theta),
                1.75 - 0.75 * np.cos(theta),
            )
        ),
        ydata=np.concatenate(
            (
                [5, 5],
                5.25 + 1.75 * np.sin(theta),
                1.75 - 1.75 * np.sin(theta),
                [3, 3, 2, 2],
                1.75 - 0.75 * np.sin(theta),
                5.25 + np.sin(theta),
            )
        ),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_h() -> LetterData:
    """Return data for letter H."""
    return LetterData(
        xdata=np.array([0, 1, 1, 2.5, 2.5, 3.5, 3.5, 2.5, 2.5, 1, 1, 0]),
        ydata=np.array([7, 7, 4, 4, 7, 7, 0, 0, 3, 3, 0, 0]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_i() -> LetterData:
    """Return data for letter I."""
    return LetterData(
        xdata=np.array([0, 3.5, 3.5, 2.25, 2.25, 3.5, 3.5, 0, 0, 1.25, 1.25, 0]),
        ydata=np.array([7, 7, 6, 6, 1, 1, 0, 0, 1, 1, 6, 6]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_j() -> LetterData:
    """Return data for letter J."""
    theta = np.linspace(0, np.pi, 20)
    return LetterData(
        xdata=np.concatenate(
            ([0, 1], 1.75 - 0.75 * np.cos(theta), [2.5, 3.5], 1.75 + 1.75 * np.cos(theta))
        ),
        ydata=np.concatenate(
            ([2, 2], 1.75 - 0.75 * np.sin(theta), [7, 7], 1.75 - 1.75 * np.sin(theta))
        ),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_k() -> LetterData:
    """Return data for letter K."""
    return LetterData(
        xdata=np.array([0, 0, 1, 1, 1.5, 2.5, 3.5, 13 / 6, 3.5, 2.5, 1, 1]),
        ydata=np.array([7, 0, 0, 2.625, 3.5, 0, 0, 14 / 3, 7, 7, 4.375, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_l() -> LetterData:
    """Return data for letter L."""
    return LetterData(
        xdata=np.array([0, 0, 3.5, 3.5, 1, 1]),
        ydata=np.array([7, 0, 0, 1, 1, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_m() -> LetterData:
    """Return data for letter M."""
    return LetterData(
        xdata=np.array([0, 0, 1, 1, 1.75, 2.5, 2.5, 3.5, 3.5, 2.5, 1.75, 1]),
        ydata=np.array([7, 0, 0, 4, 1.75, 4, 0, 0, 7, 7, 4.75, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_n() -> LetterData:
    """Return data for letter N."""
    return LetterData(
        xdata=np.array([0, 1, 2.5, 2.5, 3.5, 3.5, 2.5, 1, 1, 0]),
        ydata=np.array([7, 7, 2.8, 7, 7, 0, 0, 4.2, 0, 0]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_o() -> LetterData:
    """Return data for letter O."""
    theta = np.linspace(0, 2 * np.pi, 40)
    xdata = np.concatenate((1.75 * np.sin(theta), -0.75 * np.sin(theta)))
    ydata = np.concatenate((3.5 * np.cos(theta), 2.5 * np.cos(theta)))
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=0,
        ycenter=0,
        xdata_plot=(xdata[:40], xdata[40:]),
        ydata_plot=(ydata[:40], ydata[40:]),
    )


def capital_p() -> LetterData:
    """Return data for letter P."""
    theta = np.linspace(np.pi / 2, -np.pi / 2, 20)
    xdata = np.concatenate(([1, 0, 0], 1.5 + 2 * np.cos(theta), [1, 1], 1.5 + np.cos(theta), [1]))
    ydata = np.concatenate(([0, 0, 7], 5 + 2 * np.sin(theta), [3, 4], 5 - np.sin(theta), [6]))
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=1.75,
        ycenter=3.5,
        xdata_plot=(xdata[:24], xdata[24:]),
        ydata_plot=(ydata[:24], ydata[24:]),
    )


def capital_q() -> LetterData:
    """Return data for letter Q."""
    theta1 = np.linspace(
        -np.arcsin(1 / np.sqrt(41)) - np.arctan(4 / 5),
        2 * np.pi - np.arcsin(23 * np.sqrt(41) / 287) - np.arctan(4 / 5),
        40,
    )
    theta2 = np.linspace(
        2 * np.pi - np.arcsin(23 / np.sqrt(769)) - np.arctan(12 / 25),
        -np.arcsin(7 / np.sqrt(769)) - np.arctan(12 / 25),
        40,
    )
    xdata = np.concatenate((1.75 * np.cos(theta1), [0.75, 1.75, 0.5, -0.5], 0.75 * np.cos(theta2)))
    ydata = np.concatenate((3.5 * np.sin(theta1), [-3.5, -3.5, -1.5, -1.5], 2.5 * np.sin(theta2)))
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=0,
        ycenter=0,
        xdata_plot=(xdata[:42], xdata[42:]),
        ydata_plot=(ydata[:42], ydata[42:]),
    )


def capital_r() -> LetterData:
    """Return data for letter R."""
    theta1 = np.linspace(0, np.pi - np.arcsin(np.sqrt(10) / 20) - np.arctan(1 / 3), 20)
    theta2 = np.linspace(0, np.pi, 20)
    xdata = np.concatenate(
        ([1, 0, 0, 1], 1.5 + 2 * np.sin(theta1), [3.5, 2.5, 1.5, 1, 1], 1.5 + np.sin(theta2), [1])
    )
    ydata = np.concatenate(
        ([0, 0, 7, 7], 5 + 2 * np.cos(theta1), [0, 0, 3, 3, 4], 5 - np.cos(theta2), [6])
    )
    return LetterData(
        xdata=xdata,
        ydata=ydata,
        xcenter=1.75,
        ycenter=3.5,
        xdata_plot=(xdata[:28], xdata[28:]),
        ydata_plot=(ydata[:28], ydata[28:]),
    )


def capital_s() -> LetterData:
    """Return data for letter S."""
    theta = np.linspace(0, np.arcsin(5 / 7) + np.pi, 20)
    return LetterData(
        xdata=np.concatenate(
            (
                0.75 * np.cos(theta),
                -1.75 * np.cos(np.flip(theta)),
                -0.75 * np.cos(theta),
                1.75 * np.cos(np.flip(theta)),
            )
        ),
        ydata=np.concatenate(
            (
                1.75 + 0.75 * np.sin(theta),
                -1.75 - 1.75 * np.sin(np.flip(theta)),
                -1.75 - 0.75 * np.sin(theta),
                1.75 + 1.75 * np.sin(np.flip(theta)),
            )
        ),
        xcenter=0,
        ycenter=0,
    )


def capital_t() -> LetterData:
    """Return data for letter T."""
    return LetterData(
        xdata=np.array([0, 3.5, 3.5, 2.25, 2.25, 1.25, 1.25, 0]),
        ydata=np.array([7, 7, 6, 6, 0, 0, 6, 6]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_u() -> LetterData:
    """Return data for letter U."""
    theta1 = np.linspace(0, np.pi, 20)
    theta2 = np.linspace(0, np.arcsin(3 / 7) + np.pi / 2, 20)
    return LetterData(
        xdata=np.concatenate(
            (
                [0.75, 1.75, 1.75, 0.75],
                0.75 * np.cos(theta1),
                [-0.75, -1.75],
                -1.75 * np.cos(theta2),
            )
        ),
        ydata=np.concatenate(
            ([0, 0, 7, 7], 1.75 - 0.75 * np.sin(theta1), [7, 7], 1.75 - 1.75 * np.sin(theta2))
        ),
        xcenter=0,
        ycenter=3.5,
    )


def capital_v() -> LetterData:
    """Return data for letter V."""
    return LetterData(
        xdata=np.array([0, 1.25, 2.25, 3.5, 2.5, 1.75, 1]),
        ydata=np.array([7, 0, 0, 7, 7, 2.8, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_w() -> LetterData:
    """Return data for letter W."""
    return LetterData(
        xdata=np.array([0, 0, 1, 1.75, 2.5, 3.5, 3.5, 2.5, 2.5, 1.75, 1, 1]),
        ydata=np.array([7, 0, 0, 2.25, 0, 0, 7, 7, 3, 5.25, 3, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_x() -> LetterData:
    """Return data for letter X."""
    return LetterData(
        xdata=np.array([3.5, 2.5, 1.75, 1, 0, 1.25, 0, 1, 1.75, 2.5, 3.5, 2.25]),
        ydata=np.array([7, 7, 4.9, 7, 7, 3.5, 0, 0, 2.1, 0, 0, 3.5]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_y() -> LetterData:
    """Return data for letter Y."""
    return LetterData(
        xdata=np.array([0, 1.25, 1.25, 2.25, 2.25, 3.5, 2.5, 1.75, 1]),
        ydata=np.array([7, 3, 0, 0, 3, 7, 7, 4.6, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )


def capital_z() -> LetterData:
    """Return data for letter Z."""
    return LetterData(
        xdata=np.array([0, 0, 2.5, 0, 0, 3.5, 3.5, 1, 3.5, 3.5]),
        ydata=np.array([7, 6, 6, 1, 0, 0, 1, 1, 6, 7]),
        xcenter=1.75,
        ycenter=3.5,
    )



"""Scripts for testing arrow.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from traffic_scene_renderer import arrow

from .test_static_objects import save_fig

mpl.use("Agg")


def test_arrow() -> None:
    fig, axes = plt.subplots(1, 1)
    axes.set_xlim(-10, 10)
    axes.set_ylim(-10, 10)
    arrow(axes, [-8, -8], [-8, 8], color="r")  # Straight arrow
    arrow(axes, [-6, -6, -2, 8], [-8, 4, 8, 8], color="b")  # Arrow with 90 degrees corner
    arrow(axes, [-4, -4, 0, 0], [-8, -4, 2, 6], color="g")  # Arrow with s-bend
    theta = np.linspace(0, 3 * np.pi / 2, 50)
    arrow(axes, 5 + 4 * np.sin(theta), -4 + 4 * np.cos(theta), color="k")  # Random shape
    save_fig(fig, axes, Path("arrow") / "arrows.png")



"""Abstract class for drawing vehicles.

Author(s): Erwin de Gelder
"""

from abc import abstractmethod
from typing import Optional, Tuple

import numpy as np
from matplotlib.axes import Axes

from .options import Options
from .path_follower import PathFollower
from .static_objects import StaticObject


class MoveVehicleNoPathFollowerDefinedError(Exception):
    """Error to be raised in case a vehicle is moved without a defined PathFollower."""

    def __init__(self) -> None:
        """Description of error."""
        super().__init__("Cannot move the vehicle if no PathFollower is defined.")


class VehicleOptions(Options):
    """The default values of the options of a vehicle.

    The following list shows the options (default values within parentheses):
        length (4.5): The length of the vehicle.
        width (1.8): The width of the vehicle.
        line_width (1): The line width for plotting the vehicle.
        x_position_init (0): The initial horizontal position of the vehicle.
        y_position_init (0): The initial vertical position of the vehicle.
        angle_init (0): The initial angle of the vehicle.
        color ((0, 0.4375, 0.75)): Color of the vehicle.
        edgecolor ((0, 0, 0)): Edgecolor of the vehicle.
        layer (2): The layer in which the vehicle will be plotted.
    """

    length: float = 4.5
    width: float = 1.8
    line_width: float = 1
    x_position_init: float = 0
    y_position_init: float = 0
    angle_init: float = 0
    color: Tuple[float, float, float] = (0, 0.4375, 0.75)
    edgecolor: Tuple[float, float, float] = (0, 0, 0)
    layer: int = 2


class Vehicle(StaticObject):
    """Plot a vehicle on the scene.

    Attributes:
        options (CarOptions): All options. For a detailed description, see above.
        is_braking (bool): Whether the car is braking or not.
        axes (Axes): The axes that is used for plotting.
        path_follower (PathFollower): Optional object, used when vehicle needs to follow a path.
    """

    def __init__(
        self,
        axes: Axes,
        options: Optional[VehicleOptions] = None,
        path_follower: Optional[PathFollower] = None,
    ) -> None:
        """Instantiating a vehicle object.

        :param axes: The axes on which the vehicle is supposed to be drawn.
        :param options: Various options for the vehicle, see VehicleOptions.
        :param path_follower: Path that the vehicle needs to follow (if any).
        """
        self.options = VehicleOptions() if options is None else options
        self.is_braking = False
        self.axes = axes
        StaticObject.__init__(self, axes)

        # Plot the vehicle.
        self.plot_vehicle()

        # Update initial position if path follower is given
        self.path_follower = path_follower
        if self.path_follower is not None:
            if self.path_follower.length is None:
                self.path_follower.length = self.options.length
            self.options.x_position_init, self.options.y_position_init, self.options.angle_init = (
                self.path_follower.get_center_coordinates()
            )

        # Set initial position
        self.change_pos(
            self.options.x_position_init, self.options.y_position_init, self.options.angle_init
        )

    @abstractmethod
    def plot_vehicle(self) -> None:
        """Plot the vehicle on the axes."""

    def get_front_x(self) -> float:
        """Return the x-coordinate of the front of the vehicle.

        :return: The x-coordinate of the front of the vehicle.
        """
        return self.position.x_center + self.options.length * np.sin(self.position.angle) / 2

    def get_rear_x(self) -> float:
        """Return the x-coordinate of the rear of the vehicle.

        :return: The x-coordinate of the front of the vehicle.
        """
        return self.position.x_center - self.options.length * np.sin(self.position.angle) / 2

    def get_front_y(self) -> float:
        """Return the y-coordinate of the front of the vehicle.

        :return: The y-coordinate of the front of the vehicle.
        """
        return self.position.y_center + self.options.length * np.cos(self.position.angle) / 2

    def get_rear_y(self) -> float:
        """Return the y-coordinate of the front of the vehicle.

        :return: The y-coordinate of the front of the vehicle.
        """
        return self.position.y_center - self.options.length * np.cos(self.position.angle) / 2

    def set_front_xy(self, x_front: float, y_front: float, angle: float = 0) -> None:
        """Set position of vehicle such that its front is at (x, y) with provided angle.

        :param x_front: The x-coordinate of where the front should be.
        :param y_front: The y-coordinate of where the frnot should be.
        :param angle: The angle of the vehicle.
        """
        x_center = x_front - self.options.length * np.sin(angle) / 2
        y_center = y_front - self.options.length * np.cos(angle) / 2
        self.change_pos(x_center, y_center, angle)

    def move_vehicle(self, stepsize: float) -> None:
        """Move the vehicle a tiny bit and return new coordinates.

        This is only possible if self.path_follower is defined.
        """
        if self.path_follower is None:
            raise MoveVehicleNoPathFollowerDefinedError
        self.change_pos(*self.path_follower.move_vehicle(stepsize))

    def start_braking(self, length: float = 2) -> None:
        """Add the skid mark that shows that a vehicle is braking.

        :param length: Length of the skid mark (default=2).
        """
        if not self.is_braking:
            x_center = self.position.x_center
            y_center = self.position.y_center
            angle = self.position.angle
            self.change_pos(0, 0, 0)
            xmin = 0.3 * self.options.width
            xmax = 0.45 * self.options.width
            ymin = 0
            ymax = -0.5 * self.options.length - length
            self.fills += (
                self.axes.fill(
                    [xmin, xmin, xmax, xmax],
                    [ymin, ymax, ymax, ymin],
                    "k",
                    zorder=self.options.layer - 1,
                )[0],
                self.axes.fill(
                    [-xmin, -xmin, -xmax, -xmax],
                    [ymin, ymax, ymax, ymin],
                    "k",
                    zorder=self.options.layer - 1,
                )[0],
            )
            self.is_braking = True
            self.change_pos(x_center, y_center, angle)

    def stop_braking(self) -> None:
        """Remove the skid mark (if any) that shows that a vehicle is braking."""
        if self.is_braking:
            removes = self.fills[-2:]
            self.fills = self.fills[:-2]
            for remove in removes:
                remove.remove()
                del remove
            self.is_braking = False



"""Python library for creating schematic overviews of traffic scenes.

Author(s): Erwin de Gelder
"""

from .ambulance import Ambulance, AmbulanceOptions
from .arrow import arrow
from .bus import Bus, BusOptions
from .car import Car, CarOptions, CarType, InvalidCarTypeError
from .connection import Connection, InvalidConnectionError
from .crossing import Crossing, CrossingOptions
from .letters import Letter, LetterOptions, Letters, LettersOptions
from .options import FrozenOptionsError, Options, UnknownOptionError
from .path_follower import PathFollower
from .road_network import CrossingInfo, RoadNetwork, RoadNetworkOptions, StopLineOptions
from .static_objects import (
    Building,
    BuildingOptions,
    MaxSpeed,
    MaxSpeedOptions,
    Stripes,
    StripesOptions,
    TurnArrow,
    TurnArrowOptions,
)
from .traffic_light import NoAmberError, TrafficLight, TrafficLightOptions, TrafficLightStatus
from .truck import Truck, TruckOptions
from .vehicle import MoveVehicleNoPathFollowerDefinedError
from .vertex import Vertex, VertexOptions
from .way import IndexVertexError, Way, WayOptions



"""Constructing a vertex that is used to construct ways.

Author(s): Erwin de Gelder
"""

from typing import List, Optional, Tuple

import numpy as np

from .options import Options
from .utilities import wgs_to_utm


class VertexOptions(Options):
    """The default options for a vertex."""

    latlon: bool = False
    zonenumber: Optional[int] = None


class Vertex:
    """A vertex that can be used to construct ways.

    The following list shows the options with within parentheses the default values:
        latlon (False): Whether the provided data is in latlon format. If so, data will be
                        transformed.
        zonenumber (None): In case latlon data is provided, it will be transformed using the
                           zonenumber. If no zonenumber is provided, it will be computed based on
                           the latlon data.
    """

    def __init__(
        self, idx: int, xdata: float, ydata: float, options: Optional[VertexOptions] = None
    ) -> None:
        """Create vertex object.

        :param idx: Index of the vertex. Make sure it is unique.
        :param xdata: x-coordinate.
        :param ydata: y-coordinate.
        :param options: options, such as latlon and zonenumber.
        """
        # Define default options.
        self.options = VertexOptions() if options is None else options

        self.idx = idx
        self.xcoordinate = float(xdata)
        self.ycoordinate = float(ydata)
        if self.options.latlon:
            self.xcoordinate, self.ycoordinate = self.compute_wgs(
                force_zone_number=self.options.zonenumber
            )

    def compute_wgs(self, force_zone_number: Optional[int] = None) -> Tuple[float, float]:
        """Compute the wgs coordinates.

        :param force_zone_number: Zone number to be used. When set to none, zone is determined by
                                  latlon coordinate.
        """
        utm = wgs_to_utm(
            np.array([[self.xcoordinate, self.ycoordinate]]), force_zone_number=force_zone_number
        )
        self.options.zonenumber = utm[1]
        return utm[0][0, 0], utm[0][0, 1]

    def get_xy(self) -> List:
        """Get the (x, y)-coordinate of this vertex.

        :return: The (x, y) coordinate as a list of two numbers.
        """
        return [self.xcoordinate, self.ycoordinate]

    def __str__(self) -> str:
        """Return a string with the ID, x-coordinate, and y-coordinate of the vertex."""
        return f"Vertex[ID={self.idx:d}, x={self.xcoordinate:.2f}, y={self.ycoordinate:.2f}]"



"""Scripts for testing car.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest

from traffic_scene_renderer import (
    Car,
    CarOptions,
    CarType,
    InvalidCarTypeError,
    MoveVehicleNoPathFollowerDefinedError,
    PathFollower,
)

from .test_static_objects import save_fig

mpl.use("Agg")


def test_car_creation_without_options() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-3, 3)
    axes.set_ylim(-3, 3)
    Car(axes)
    save_fig(fig, axes, Path("car") / "default_car.png", 3)


def test_cars_transparent() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-5, 5)
    axes.set_ylim(-3, 3)
    for i, car_type in enumerate(
        (CarType.VEHICLE, CarType.STATION_WAGON, CarType.PICKUP_TRUCK, CarType.SEDAN)
    ):
        car = Car(axes, CarOptions(fill=False, icar=car_type))
        car.change_pos(-3.75 + 2.5 * i, 0)
    save_fig(fig, axes, Path("car") / "transparent_cars.png", 5)


def test_invalid_car_type_error() -> None:
    fig, axes = plt.subplots()
    try:
        Car(axes, CarOptions(fill=False, icar=5))
    except InvalidCarTypeError:
        pass
    else:
        pytest.fail("InvalidCarTypeError should be raised.")
    finally:
        plt.close(fig)


def test_car_change_color() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-3, 3)
    axes.set_ylim(-3, 3)
    car = Car(axes)
    car.change_color(face_color=(1, 0, 0), edge_color=(0.4, 0, 0))
    save_fig(fig, axes, Path("car") / "red_car.png", 3)


def test_car_get_coordinates() -> None:
    fig, axes = plt.subplots()
    car = Car(axes, CarOptions(length=2))
    assert car.get_front_x() == 0.0
    assert car.get_front_y() == 1.0
    assert car.get_rear_x() == 0.0
    assert car.get_rear_y() == -1.0
    plt.close(fig)


def test_car_set_front_xy() -> None:
    fig, axes = plt.subplots()
    car = Car(axes, CarOptions(length=2))
    car.set_front_xy(1, 0, np.pi / 2)
    assert (car.get_front_x(), car.get_front_y()) == (1.0, 0.0)
    plt.close(fig)


def test_car_braking() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-3, 3)
    axes.set_ylim(-3, 3)
    car_braking = Car(axes)
    car_braking.change_pos(-1.5, 0)
    car_braking.start_braking()

    car_not_braking = Car(axes)
    car_not_braking.change_pos(1.5, 0)
    car_not_braking.start_braking()
    car_not_braking.stop_braking()

    axes.text(-1.5, 2.5, "Braking", horizontalalignment="center")
    axes.text(1.5, 2.5, "Not braking", horizontalalignment="center")
    save_fig(fig, axes, Path("car") / "braking_car.png", 3)


def test_move_vehicle_no_path_follower_defined_error() -> None:
    fig, axes = plt.subplots()
    car = Car(axes)
    try:
        car.move_vehicle(1)
    except MoveVehicleNoPathFollowerDefinedError:
        pass
    else:
        pytest.fail("MoveVehicleNoPathFollowerDefinedError should be raised.")
    finally:
        plt.close(fig)


def test_car_moving() -> None:
    fig, axes = plt.subplots()
    path_follower = PathFollower(np.array([0, 0]), np.array([0, 10]))
    car = Car(axes, path_follower=path_follower)
    car.move_vehicle(car.options.length)
    assert car.get_rear_y() == car.options.length
    plt.close(fig)



"""Scripts for testing ambulance.py.

Author(s): Erwin de Gelder
"""

from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt

from traffic_scene_renderer import Ambulance

from .test_static_objects import save_fig

mpl.use("Agg")


def test_ambulance_creation() -> None:
    fig, axes = plt.subplots()
    axes.set_xlim(-2, 2)
    axes.set_ylim(-4, 4)
    Ambulance(axes)
    save_fig(fig, axes, Path("ambulance") / "standard_ambulance.png", 5)
